"""Module containing classes abstracting datasets
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import abc
import itertools
import logging
import pathlib
import re
import shelve
import string
import shutil
import tempfile
import datetime
import sys
import collections
import warnings

import numpy
import numpy.lib.arraysetops
import numpy.lib.recfunctions

import netCDF4

import xarray

from .. import config
from .. import utils
from .. import math as tpmath
from ..physics.units import em
from ..constants import MiB

from . import filters

logger = logging.getLogger(__name__)

try:
    import progressbar
except ImportError:
    progressbar = None

class GranuleLocatorError(Exception):
    """Problem locating granules.
    """

class DataFileError(Exception):
    """Superclass for any datafile problems

    Upon reading a large amounts of data, some files will contain
    problems.  When processing a year of data, we don't want to fail
    entirely when one orbit file fails.  But if there is a real bug
    somewhere, we probably do want to fail.  Therefore, the typhon dataset
    framework is optionally resilient to datafile related errors, but only
    if those derive from DataFileError.

    Therefore, users implementing their own reading routine may wish to
    catch errors arising from corrupted data, and raise an exception
    derived from DataFileError instead.  That way, typhon knows that a
    problem is due to corrupted data and not due to a bug.
    """

class InvalidFileError(DataFileError, ValueError):
    """Raised when the requested information cannot be obtained from the file

    See DataFileError for more information.
    """

class InvalidDataError(DataFileError):
    """Raised when data is not how it should be.

    See DataFileError for more information.
    """

all_datasets = {}

class Dataset(metaclass=abc.ABCMeta):
    """Represents a dataset.

    This is an abstract class.  More specific subclasses are
    SingleFileDataset and MultiFileDataset.
    
    To add a dataset, subclass one of the subclasses of Dataset, such as
    MultiFileDataset, and implement the abstract methods.

    Dataset objects have a limited number of attributes.  To limit the
    occurence of bugs, dynamically setting non-pre-existing attributes is
    limited.  Attributes can be set either by passing keyword arguments
    when creating the object, or by setting the appropriate field in your
    typhon configuration file (such as .typhonrc).  The configuration
    section will correspond to the object name, the key to the attribute,
    and the value to the value assigned to the attribute.  See also
    :mod:`typhon.config`.

    Attributes:

        start_date (datetime.datetime or numpy.datetime64):
            Starting date for dataset.  May be used to search through ALL
            granules.  WARNING!  If this is set at a time t_0 before the
            actual first measurement t_1, then the collocation algorith (see
            CollocatedDataset) will conclude that there are 0 collocations
            in [t_0, t_1], and will not realise if data in [t_0, t_1] are
            actually added later!
        end_date (datetime.datetime or numpy.datetime64):
            Similar to start_date, but for ending.
        name (str):
            Name for the dataset.  Used to make sure there is only a
            single dataset with the same name for any particular dataset.
            If a dataset is initiated with a pre-exisitng name, the
            previous product is called.
        aliases (Mapping[str, str]):
            Aliases for field.  Dictionary can be useful if you want to
            programmatically loop through the same field for many different
            datasets, but they are named differently.  For example, an alias
            could be "ch4_profile".
        unique_fields (Container[str]):
            Set of fields that make any individual measurement unique.  For
            example, the default value is {"time", "lat", "lon"}.
        related (Mapping[str, Dataset]):
            Dictionary whose keys may refer to other datasets with related
            information, such as DMPs or flags.

    """

    _instances = None

    start_date = None
    end_date = None
    name = ""
    section = ""
    aliases = {}
    time_field = "time"
    unique_fields = {"time", "lat", "lon"}
    mandatory_fields = None # fields that reader relies upon
    related = {}
    maxsize = 10000*MiB
    my_pseudo_fields = None
    read_returns = "ndarray"
    default_orbit_filters = []

    # in case of xarray returns, if concatenation is NOT along time
    # coordinate, this should be set to something else
    concat_coor = None

    # Make singleton: there is no point in having multiple copies of a
    # dataset-object around when both relate to the same dataset and
    # class.
    def __new__(cls, name=None, **kwargs):
        name = name or cls.name or cls.__name__
        if cls._instances is None:
            cls._instances = {}
        # make sure subclasses aren't confused for their parents
        if not cls in cls._instances:
            cls._instances[cls] = {}
        if name in cls._instances[cls]:
            return cls._instances[cls][name]
        self = super().__new__(cls)
        cls._instances[cls][name] = self
        if not cls in all_datasets:
            all_datasets[cls] = {}
        all_datasets[cls][name] = self
        return self # __init__ takes care of the rest

    # through __new__, I make sure there's only a single dataset
    # object per class with the same name.  However, __init__ will get
    # called again despite __new__ returning an older self.  Make sure
    # we don't get hurt by that.
    def __init__(self, **kwargs):
        """Initialise a Dataset object.

        All keyword arguments will be translated into attributes.
        Does not take positional arguments.

        Note that if you create a dataset with a name that already exists,
        the existing object is returned, but __init__ is still called
        (Python does this, see
        https://docs.python.org/3.7/reference/datamodel.html#object.__new__).
        """
        self.mandatory_fields = set()
        for (k, v) in kwargs.items():
            setattr(self, k, v)
        self.setlocal()
        if self.my_pseudo_fields is None:
            self.my_pseudo_fields = collections.OrderedDict()

    def __setattr__(self, k, v):
        if hasattr(self, k) or hasattr(type(self), k):
            super().__setattr__(k, v)
        else:
            raise AttributeError("Unknown attribute: {}. ".format(k))

    def setlocal(self):
        """Set local attributes, from config or otherwise.

        """
        if self.section in config.conf:
            for k in config.conf[self.section]:
                setattr(self, k, config.conf[self.section][k])

    @abc.abstractmethod
    def find_granules(self, start=datetime.datetime.min,
                                  end=datetime.datetime.max,
                    include_last_before=False):
        """Loop through all granules for indicated period.

        This is a generator that will loop through all granules from
        `start` to `end`, inclusive.

        See also: `find_granules_sorted`

        Arguments:

            start (datetime.datetime): Start
                Starting datetime.  When omitted, start at complete
                beginning of dataset.

            end (datetime.datetime): End
                End datetime.  When omitted, continue to end of dataset.
                Last granule will start before this datetime, but contents
                may continue beyond it.

            include_last_before (bool): Be inclusive
                When True, also return the last granule /before/ start, so
                that a reader is sure to include all data in the covered
                period.  When False, the first granule yielded is the
                first granule starting after start.
                 
        Yields:
            pathlib.Path objects for all files in dataset.  Sorting is not
            guaranteed; if you need guaranteed sorting, use
            `find_granules_sorted`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def find_granules_sorted(self, start=None, end=None):
        """Yield all granules sorted by starting time then ending time.

        For details, see `find_granules`.
        """
        raise NotImplementedError()
    
    @abc.abstractmethod
    def find_most_recent_granule_before(self, instant,
            **locator_args):
        """Find granule covering instant

        Find granule started most recently before `instant`.

        Arguments:

            instant (datetime.datetime): Time to search for
                datetime for which a granule is sought
                beginning of dataset.

            **locator_args:
                Any other keyword arguments that the particular dataset
                needs.  Commonly, `satname` is needed.
                 
        Returns:
            pathlib.Path object for sought granule.

        """
        ...

    # Cannot use functools.lru_cache because it cannot handle mutable
    # results or arguments
    @utils.cache.mutable_cache(maxsize=10)
    def read_period(self, start=None,
                          end=None,
                          onerror="skip",
                          fields="all",
                          pseudo_fields=None,
                          sorted=True,
                          locator_args=None,
                          reader_args=None,
                          limits=None,
                          simple_filters=(),
                          orbit_filters=None,
                          enforce_no_duplicates=True,
                          excs=(DataFileError, filters.FilterError)):
        """Read all granules between start and end, in bulk.

        Arguments:

            start (datetime.datetime): Start
                Starting datetime.  When omitted, start at complete
                beginning of dataset.

            end (datetime.datetime): End
                End datetime.  When omitted, continue to end of dataset.
                Last granule will start before this datetime, but contents
                may continue beyond it.  Can also be a datetime.timedelta,
                which will then be interpreted as a length after start.
 
            onerror (str): What to do on errors
                When reading many files, some files may have problems.  If
                onerror is set to "skip", files with errors are skipped
                and reading continues with the next file.  This is the
                default behaviour.  If onerror is set to "raise", the
                method will reraise the original exception as soon as a
                problem occurs.

            fields (Iterable[str] or str): What fields to return
                What fields to read from dataset.  Either a collection
                of strings corresponding to fields to read, or the str
                "all" (default), which means read all fields.

            pseudo_fields (Mapping[str, function]): See documentation for
                self.read.

            sorted (bool): Should the granules be read in sorted order?
                Defaults to true.  NB: does not currently guarantee that
                the actual results are sorted; this is up to the
                individual reading routines!

            locator_args (dict): Extra keyword arguments passed on to
                granule finding routines.

            reader_args (dict): Extra keyword arguments to be passed on to
                reading routine.  Since these differ per type, those are
                probably documented in the class docstring.

            limits (dict): Limitations to apply to each granule.  For the
                exact format, see `:func:typhon.math.array.limit_ndarray`.

            simple_filters (container/iterable): collection of functions to be
                applied for filtering.  Must take ndarray input, must give
                ndarray output.  For more advanced filtering, pass
                instances of implementations of filters.OrbitFilter to
                orbit_filters.
    
            orbit_filters (Sequence[filters.OrbitFilter]): more advanced
                filters to apply for each orbit or after all orbits.
                Those are instances of filters.OrbitFilter and thus have
                state, which can be useful for e.g. overlap checking.  For
                each of those, the system will call .reset() before the
                first orbit is read, .filter(...) after each orbit, and
                .finalise() at the end.

            enforce_no_duplicates

            excs (Sequence[Exception]): what exceptions to try and catch
                after every read.  Use with onorrer.

        Returns:
            
            Masked array containing all data in period.  Invalid data may
            or may not be masked, depending on the behaviour of the
            reading routine implemented by the subclass.
        """

        if locator_args is None:
            locator_args = {}

        if reader_args is None:
            reader_args = {}

        if limits is None:
            limits = {}

        if orbit_filters is None:
            orbit_filters = self.default_orbit_filters

        start = start or self.start_date
        end = end or self.end_date
        if isinstance(end, datetime.timedelta):
            end = start + end

        finder = self.find_granules_sorted if sorted else self.find_granules
        logger.info("Reading {self.name:s} {locator_args!s} "
                     "for period {start:%Y-%m-%d %H:%M:%S} "
                     " – {end:%Y-%m-%d %H:%M:%S}".format(**vars()))

        dobar = sorted and progressbar and sys.stdout.isatty()
        if dobar:
            bar = progressbar.ProgressBar(max_value=1,
                    widgets=[progressbar.Bar("=", "[", "]"), " ",
                             progressbar.Percentage(),
                             ' (', progressbar.ETA(), ') '])
            bar.start()
            bar.update(0)
        else:
            logger.info("Psst!  If you install the progressbar2 package, "
                "you will get a fancy progressbar!")
        anygood = False
        arr = None
        N = 0
        time = self.time_field
        for of in orbit_filters:
            of.reset()
        # NB: https://stackoverflow.com/a/37585628/974555
        extra_filter_args = collections.ChainMap(*(of.args_to_reader
            for of in orbit_filters))

        overlap_filters = [
            f for f in orbit_filters if isinstance(f, filters.OverlapFilter)]
        late = overlap_filters[0].late if len(overlap_filters) > 0 else True
        if len(overlap_filters) > 1:
            raise ValueError("Found {:d} overlap filters: {!s}".format(
                len(overlap_filters), overlap_filters))
        for (g_start, gran) in finder(start, end, return_time=True, 
                                      include_last_before=True,
                                      **locator_args):
            try:
                (cont, extra) = self.read(str(gran), fields=fields,
                    pseudo_fields=pseudo_fields, **reader_args,
                    **extra_filter_args)
                for of in orbit_filters:
                    cont = of.filter(cont, **extra)
                # FIXME: handle cases where very few scanlines are left
                # after filtering.  We may find errors downstream if there
                # are very few scanlines left.
                if (not late and
                    cont[time].size > 0):

                    if (enforce_no_duplicates and sorted and
                            arr is not None and
                            N>0 and (cont[time][0] <= latest)):
                        if isinstance(latest, xarray.DataArray):
                            latest = latest.values.astype("M8[ms]")
                        if isinstance(cont[time][0], xarray.DataArray):
                            conttime = cont[time][0].values.astype("M8[ms]")
                        else:
                            conttime = cont[time][0]
                        raise InvalidDataError(
                            "Reading routine for {!s} returned data starting "
                            "{:%Y-%m-%d %H:%M:%S}, which precedes last entry "
                            "for preceding granule at {:%Y-%m-%d %H:%M:%S}. "
                            "As data are supposed to be sorted in time, this "
                            "probably means duplicate removal is not working "
                            "as it should, or you are using an implementation "
                            "that deliberately leaves duplicates in, and you "
                            "should be passing enforce_no_duplicates=False to "
                            "read_period.".format(gran,
                                conttime.astype(datetime.datetime),
                                latest.astype(datetime.datetime)))

                    # NB: when datasets erroneously contain duplicate coordinates
                    # (I'm looking at you, FIDUCEO/FCDR_HIRS#159!), this
                    # comparison will cause a failure in xarray.  In this case,
                    # compare values instead.
                    arrrr = cont[time]
                    if isinstance(arrrr, xarray.DataArray):
                        arrrr = arrrr.values
                    if not (arrrr[1:] >= arrrr[:-1]).all():
                        raise InvalidDataError("Reader for {!s} returned data "
                            "with unsorted time.  This must be fixed.".format(
                                f))

                cont = self._apply_limits_and_filters(cont, limits, simple_filters)

            except excs as exc:
                if onerror == "skip": # fields that reader relies upon
                    logger.error("Can not read file {}: {}".format(
                        gran, exc.args[0]))
                    continue
                else:
                    raise
            else:
                (arr, N, latest) = self._add_gran_to_data(arr, cont, N, start,
                                     end, g_start)
#                contents.append(
#                    cont[(cont[time]<=end)&(cont[time]>=start)])
                if N > 0:
                    anygood = True
            if dobar:
                bar.update(max((g_start-start) / (end-start),0))
        if dobar:
            bar.update(1)
            bar.finish()
        if anygood:
            # NB: filter finalise before or after dataset finalise?
            #
            arr = self._finalise_arr(arr, N)
            for of in orbit_filters:
                arr = of.finalise(arr)

            if "flags" in self.related:
                arr = self.flag(arr)
                    
            return arr
        else:
            raise DataFileError("Can not find any valid data!")

    def _apply_limits_and_filters(self, cont, limits, simple_filters):
        if isinstance(cont, xarray.Dataset):
            if len(limits)>0:
                raise NotImplementedError( 
                    "limits not implemented on xarray datasets")
            oldsize = cont[self.time_field].size
            for f in simple_filters:
                cont = f(cont)
            logger.debug("Filters reduced number from "
                "{:d} to {:d}".format(oldsize, cont[self.time_field].size))
            return cont
        oldsize = cont.size
        cont = tpmath.array.limit_ndarray(cont, limits)
        for f in simple_filters:
            cont = f(cont)
        if cont.size < oldsize:
            logger.debug("Applying limitations, reducing "
                "{:d} to {:d}".format(oldsize, cont.size))
        return cont

    def _add_gran_to_data(self, arr, cont, N, start, end, g_start):
        # first ensure we only add the segment we actually want
        if isinstance(cont, xarray.Dataset):
            tm = cont[self.time_field].values.astype("M8[ms]")
            cont = cont.loc[dict.fromkeys(
                utils.get_time_coordinates(cont)&cont.dims.keys(),
                slice(start, end))]
        else:
            cont = cont[(cont[self.time_field]<end)&(cont[self.time_field]>=start)]
        if isinstance(cont, xarray.Dataset):
            # xarray reads lazily, but we want to merge all granules into
            # a single object at the end, so have to load it into memory
            # anyway; rather do it now than all at once later, so that we
            # can keep track of how fast things are going
            if cont[self.time_field].size > 0: # avoid https://github.com/pydata/xarray/issues/1329
                cont.load()
            if arr is None:
                arr = [cont]
                N = cont[self.time_field].size
            else:
                arr.append(cont)
                N += cont[self.time_field].size
            for i in range(len(arr)):
                if arr[-i][self.time_field].size > 0:
                    latest = arr[-i][self.time_field][-1]
                    break
            else:
                latest = numpy.datetime64(0, "ms")
        else:
            if arr is None:
                arr = cont
                N = cont.size
            else:
                if (N+cont.size) > arr.size: # need to allocate more
                    frac_done = max((g_start-start) / (end-start),0)
                    # suppose all future files have on average the
                    # same size?  Until we have reached 10%, simply
                    # double every time.  After that, extrapolate more
                    # cleverly. 
                    newsize = (int((N+cont[self.time_field].size)//frac_done)
                                if frac_done > 0.1
                                else (N+cont[self.time_field].size) * 2)
                    arr = self._ensure_large_enough(arr, cont, N, newsize,
                        frac_done)
                self._add_cont_to_arr(arr, N, cont)
                N += cont[self.time_field].size
            if arr[self.time_field].size > 0:
                latest = arr[self.time_field][N-1]
            else:
                latest = numpy.datetime64(0, "ms")
        return arr, N, latest

    def _ensure_large_enough(self, arr, cont, N, newsize, frac_done):
        """Allocate new space while adding gran to data
        
        Helper for _add_gran_to_data, part of the read_period family of
        helpers.  Does NOT add cont to arr!"""
        
        if isinstance(cont, xarray.Dataset):
            raise NotImplementedError("Not used for xarray datasets. "
                "But see version history at "
                "https://arts.mi.uni-hamburg.de/trac/rt/browser/typhon/trunk/typhon/datasets/dataset.py?rev=10396#L462 "
                "if there is a wish to reimplemented!")
        else:
            if newsize * arr.itemsize > self.maxsize:
                raise MemoryError("This dataset is too large "
                    "for typhons little mind.  Continuing might "
                    "ultimately need {:,.0f} MiB of RAM.  This exceeds my "
                    "maximum (self.maxsize) of {:,.0f} MiB. "
                    "Sorry! ".format(
                        newsize*arr.itemsize/MiB,
                        self.maxsize/MiB))
            logger.debug(
                "New size ({:d} items, {:,.0f} MiB) would exceed allocated "
                "size ({:d} items, {:,.0f} MiB).  I'm {:.3%} "
                "through.  Allocating new: {:d} items, {:,.0f} "
                "MiB.  New size: {:d} items, {:,.0f} "
                "MiB.".format(N+cont.size,
                    (cont.nbytes+arr.nbytes)/MiB,
                    arr.size, arr.nbytes/MiB, frac_done,
                    newsize-arr.size, (newsize-arr.size)*arr.itemsize/MiB,
                    newsize, newsize*arr.itemsize/MiB))
            mod = (numpy.ma if hasattr(arr, "mask") else numpy)
            arr = mod.concatenate(
                (arr, mod.zeros(dtype=arr.dtype, shape=newsize-arr.size)))
        return arr

    def _add_cont_to_arr(self, arr, N, cont):
        """Changes arr in-situ, does not return"""
        if isinstance(cont, xarray.Dataset):
            # we should already know it's large enough
            # for arr[self.time_field] I start at N
            # for the other time coordinates at the relative "speed" they
            # are behind N
            # but this is not guaranteed to be regular so I would need to
            # keep trac of each individually, or inspect it on-the-fly
            # this approximation may be good enough for pre-allocation
            # (which is approximate anyway), when actually storing we need
            # to do a better job… for each time coordinate, check when it
            # “dies”
            raise NotImplementedError("This is not used for xarrays. "
                "But see comment in source-code for some thoughts.")
        else:
            arr[N:(N+cont.size)] = cont
            #arr = self._finalise_arr(arr, N)

    def _finalise_arr(self, arr, N):
        if isinstance(arr, list):
            logger.debug("Concatenating {N:d} DataArrays...".format(N=N))
            if self.concat_coor is None:
                return utils.concat_each_time_coordinate(*arr)
            else:
                return xarray.concat(arr, dim=self.concat_coor)
            logger.debug("Done!")
        else:
            return self._correct_overallocation(arr, N)

    @staticmethod
    def _correct_overallocation(arr, N):
        if isinstance(arr, xarray.Dataset):
            raise RuntimeError("We shouldn't be here.  Ever.")
        logger.debug("Correcting overallocation ({:d}->{:d})".format(
            arr.size, N))
        return arr[:N]

    @abc.abstractmethod
    def _read(self, f, fields="all"):
        """Read granule in file, low-level.

        To be implemented by subclass.  Do not call this method; call
        :func:`Dataset.read` instead.

        For documentation, see :func:`Dataset.read`.
        """

        raise NotImplementedError()

    # Cannot use functools.lru_cache because it cannot handle mutable
    # results or arguments
    @utils.cache.mutable_cache(maxsize=10)
    def read(self, f=None, fields="all", pseudo_fields=None, **kwargs):
        """Read granule in file and do some other fixes

        Shall return an ndarray with at least the fields lat, lon, time.

        Arguments:

            f (str): String containing path to file to read.

            fields (Iterable[str] or str): What fields to return.
                See :func:`Dataset.read_period` for details.

            pseudo_fields (Mapping[str, function]): Additional fields to
                calculate on-the-fly after every read.  That may be more
                attractive from a memory point of view.  In this mapping,
                the keys will be names added to the dtype of the returned
                ndarray (at the top level).  The values are functions.
                Each function must take four arguments:
                    
                    content [ndarray or xarray.DataArray]

                        Content of file.

                    D [Mapping]

                        Output of earlier-executed data-fields.

                    header [ndarray or xarry.DataArray]

                        Header of file.

                    fn [string-like]

                        Name of file.

                and return an ndarray or xarray.DataArray to be merged
                with content.  Dictionary are ordered now, make sure to
                pass the correct order.

            Any further keyword arguments are passed on to the particular
            reading routine.  For details, please refer to the docstring
            for the class.

        Returns:

            Masked array containing data in file with selected fields.
        """
        if isinstance(f, pathlib.PurePath):
            f = str(f)
        if pseudo_fields is None:
            pseudo_fields = collections.OrderedDict()
        # expand pseudo_fields with items from my_pseudo_fields not
        # already there, and also process dependent fields
        if fields != "all":
            fields = list(fields)
            for (k, (deps, v, cond)) in self.my_pseudo_fields.items():
                if k in fields:
                    for d in deps:
                        if (d not in fields and
                                all(kwargs.get(condfn) in condval for (condfn,
                                condval) in cond.items())):
                            fields.append(d)
            for mandatory in self.mandatory_fields:
                if mandatory not in fields:
                    fields.append(mandatory)
        for (k, (deps, v, cond)) in self.my_pseudo_fields.items():
            if (fields=="all" or 
                    k in fields and k not in pseudo_fields) and (
                    all(kwargs.get(condfn) in condval
                            for (condfn, condval) in cond.items())):
                pseudo_fields[k] = v
        logger.debug("Reading {:s}".format(f))
        # should not pass pseudo_fields on to reader, it will get confused
        fields = (fields if fields == "all" else
                  [f for f in fields if not f in pseudo_fields])
        (M, extra) = (self._read(f, fields=fields, **kwargs)
                if f is not None
                else self._read(fields=fields, **kwargs))
        M = self._add_pseudo_fields(M, pseudo_fields, extra, f)
        return (M, extra)

    def _add_pseudo_fields(self, M, pseudo_fields, extra, f):
        D = collections.OrderedDict()
        for (k, fnc) in pseudo_fields.items():
            try:
                D[k] = fnc(M, D, extra, f)
            except TypeError as exc:
                if "positional argument" in exc.args[0]:
                    # backward compatibility
                    D[k] = fnc(M)
                else:
                    raise
        if D != {}:
            if self.read_returns == "ndarray":
                mod = (numpy.ma if hasattr(M, "mask") else numpy)
                newM = mod.zeros(shape=M.shape,
                    dtype=M.dtype.descr + [(k, v.dtype, v.shape[1:]) for (k,v) in D.items()])
                for k in M.dtype.names:
                    newM[k] = M[k]
            elif self.read_returns == "xarray":
                newM = M
            else:
                raise ValueError("read_returns must be ndarray or xarray")

            for (k, v) in D.items():
                newM[k] = v
                
            M = newM
        return M

    def __str__(self):
        return "<{self.__class__.__name__:s}:{self.name:s}>".format(self=self)

#   Leave disk-memoisation out of typhon until dependency on joblib has
#   been decided.
#
#    @tools.mark_for_disk_cache(
#        process=dict(
#            my_data=lambda x: x.view(dtype="i1")))
    def combine(self, my_data, other_obj, other_data=None, other_args=None, trans=None,
                timetol=numpy.timedelta64(1, 's'), time_name="time"):
        """Combine with data from other dataset.

        Combine a set of measurements from this dataset with another
        dataset, where each individual measurement correspond to exactly
        one from the other one, as identified by time/lat/lon, orbitid, or
        measurument id, or other characteristics.  The object attribute
        unique_fields determines how those are found.

        The other dataset may contain flags, DMPs, or different
        information altogether.

        Arguments:
        
            my_data (ndarray): Data for self.
                A (masked) array with a dtype such as returned from
                `self.read <Dataset.read>`.

            other_obj (Dataset): Dataset to match
                Object from a Dataset subclass from which to find matching
                data.

            other_data (ndarray): Data for other.  Optional.
                Optionally, pass data for other object.  If not provided
                or None, this will be read using other_obj.

            other_args (dict): Keyword arguments passed to
                other_obj.read_period.  May need to contain things like
                {"locator_args": {"satname": "noaa18"}}

            trans (collections.OrderedDict): Dictionary of what field in `my_data`
                corresponds to what field in `other_data`.  Optional; by
                default, merges self.unique_fields and
                other_obj.unique_fields, and assumes names between the two
                are identical.  Order is relevant for optimal recursive
                bisection search for matches, which is to be implemented.
    
            timetol (timedelta64): For datetime types, `isclose` does not
                work (https://github.com/numpy/numpy/issues/5610).  User
                must pass an explicit tolerance, defaulting to 1 second.

            time_name (str): Name for my time field.  Defaults to "time".
                Used to determine the period to read from the other
                dataset.

        Returns:

            Masked ndarray of same size as `my_data` and same `dtype` as
            returned by `other_obj.read`.

        TODO: Allow user to pass already-read data from other dataset.
        """

        if trans is None:
            flds = list(self.unique_fields & other_obj.unique_fields)
            trans = collections.OrderedDict(zip(flds, flds))

        if other_args is None:
            other_args = {}

        if self.read_returns == "ndarray":
            if other_data is None:
                first = my_data[time_name].min().astype(datetime.datetime)
                last = my_data[time_name].max().astype(datetime.datetime)

        elif self.read_returns == "xarray":
            if other_data is None:
                first = my_data[time_name].min().values.astype("M8[ms]"
                    ).astype(datetime.datetime)
                last = my_data[time_name].max().values.astype("M8[ms]"
                    ).astype(datetime.datetime)
        else:
            raise ValueError("read_returns must be ndarray or xarray")

        if other_data is None:
            other_data = other_obj.read_period(first, last,
                            **other_args)
        (my_prim, other_prim) = next(iter(trans.items()))

        if self.read_returns == "ndarray":
            # my_data needs to be sorted
            try:
                my_data._fill_value
            except AttributeError:
                my_data.sort(order=my_prim, kind="heapsort")
            else:
                # see 
                # https://github.com/numpy/numpy/issues/8069
                my_data.sort(order=my_prim, kind="heapsort",
                                    fill_value=my_data._fill_value)

        elif self.read_returns == "xarray":

            # ensure sorted along time dimension
            if not len(my_data[my_prim].dims) == len(other_data[other_prim].dims) == 1:
                raise ValueError(
                    "Must use 1-D index for finding combinations, "
                    "{:s} has {:d} dimensions: {!s}, "
                    "{:s} has {:d} dimensions: {!s}.".format(
                        my_prim, len(my_data[my_prim].dims),
                        my_data[my_prim].dims,
                        other_prim, len(other_data[other_prim].dims),
                        other_data[other_prim].dims))
            my_dim = my_data[my_prim].dims[0]
            other_dim = other_data[other_prim].dims[0]
            if not (my_data[my_prim][1:] >= my_data[my_prim][:-1]).all():
                warnings.warn("Primary/reference instrument data should "
                    "be sorted in time to find matching data in other set. "
                    f"Field {my_prim:s} in {self.name:s} is not sorted in "
                    "time.  Note that this problem may be unavoidable if "
                    "we are tasked to search for additional information "
                    "for both pairs of collocations, and times are sorted "
                    "according to one but not the other.  Therefore, I "
                    "will force the sorting for now, but I'm not entirely "
                    "sure this cannot lead to problems later.  You have "
                    "been warned.", UserWarning)
                my_data = my_data.loc[
                    {my_data[my_prim].dims[0]:
                    numpy.argsort(my_data[my_prim])}]

        # through interpolation, find times in `other` closest to times in
        # myself; hopefully that means the difference is zero
        if issubclass(my_data[my_prim].dtype.type, numpy.datetime64):
            x = other_data[other_prim].astype("<M8[ms]").astype("f8")
            xx = my_data[my_prim].astype("<M8[ms]").astype("f8")
        else:
            x = other_data[other_prim]
            xx = my_data[my_prim]

        if self.read_returns == "ndarray":
            ii = numpy.interp(xx, x,
                numpy.arange(other_data[other_prim].shape[0])).round().astype(
                    numpy.uint64)
        else:
            if not other_data[other_prim].dtype.kind=="M":
                raise ValueError("When finding combinations based "
                    "on xarray datasets, I can only do so based on "
                    "time coordinates, but {:s} has dtype {!s}.".format(
                        other_prim, other_data[other_prim].dtype))
            for (k, v) in other_data.data_vars.items():
                if v.dtype.kind == "M":
                    raise ValueError("Found time data variable "
                        "that is not a coordinate, I'm confused, "
                        "should I interpolate it?  Assign as a coordinate "
                        "to be sure!")
            ii = dict.fromkeys(utils.get_time_dimensions(other_data))
            for k in ii.keys():
                x = other_data[k].astype("<M8[ms]").astype("f8")
                ii[k] = numpy.interp(xx, x,
                    numpy.arange(x.shape[0])).round().astype(numpy.int64)

#        if self.read_returns == "ndarray":
#            other_combi = other_data[ii]
#            # now go through fields to see that there is an actual match
#        elif self.read_returns == "xarray":
        other_combi = other_data[ii]

        if self.read_returns == "ndarray":
            near = numpy.all([(numpy.isclose(my_data[f_my], other_combi[f_oth])
                    if issubclass(my_data[f_my].dtype.type,
                                  numpy.inexact) else
                    abs(my_data[f_my] - other_combi[f_oth]) < timetol
                    if issubclass(my_data[f_my].dtype.type,
                                  numpy.datetime64) else
                    my_data[f_my] == other_combi[f_oth])
                        for (f_my, f_oth) in trans.items()], 0)
        elif self.read_returns == "xarray":
            near = numpy.all([
                (numpy.isclose(my_data[f_my].values,
                               other_combi[f_oth].values)
                    if issubclass(my_data[f_my].dtype.type,
                                  numpy.inexact) else
                    abs(my_data[f_my].values -
                        other_combi[f_oth].values) < timetol
                    if issubclass(my_data[f_my].dtype.type,
                                  numpy.datetime64) else
                    my_data[f_my].values == other_combi[f_oth].values)
                        for (f_my, f_oth) in trans.items()], 0)

        if not near.any():
            # check time coverage
            first_found = other_data["time"].min().values.astype("M8[ms]").astype(datetime.datetime)
            last_found = other_data["time"].max().values.astype("M8[ms]").astype(datetime.datetime)
            if (abs(first_found - first) > timetol and
                abs(last_found - last) > timetol):
                raise ValueError(f"Primary covers "
                    f"{first:%Y-%m-%d %H:%M:%S}–{last:%Y-%m-%d %H:%M:%S}, "
                    "but secondary only covers "
                    f"{first_found:%Y-%m-%d %H:%M:%S}–{last_found:%Y-%m-%d %H:%M:%S}. "
                    "Finding secondary fails.  Perhaps the "
                    "secondary is coming from a derived dataset compared "
                    "to where the original collocations were made, and the "
                    "derived dataset is temporarily absent.")
            else:
                raise ValueError("Did not find any secondaries!  Are times off?")

        if not near.all():
            logger.warn("Only {:d}/{:d} ({:%}) of secondaries found".format(
                near.sum(), near.size, near.sum()/near.size))

        if self.read_returns == "ndarray":
            try:
                other_combi.mask[~near] = numpy.ones_like(other_combi.mask[~near])
            except AttributeError: # not a MaskedArray yet
                other_combi = numpy.ma.masked_where(~near, other_combi)
        elif self.read_returns == "xarray":
            vars_with_dim = [k for k in other_combi.data_vars.keys()
                             if other_dim in other_combi[k].dims]
            for v in vars_with_dim:
                other_combi[v][{"time": ~near}] = numpy.nan
            
        return other_combi

    def get_additional_field(self, M, fld):
        """Get additional field.

        Get field from other dataset, original objects, or otherwise.
        To be implemented by subclass implementations.

        Exact fields depend on subclass.

        Arguments:

            M (ndarray): ndarray with existing data
                A (masked) array with a dtype such as returned from
                `self.read <Dataset.read>`.

            fld (str): Additional field to read from original data

        Returns:
            
            ndarray with fields of M + fld.
        """
        raise NotImplementedError("Must be implemented by child-class")

    def as_xarray_dataset(self):
        raise NotImplementedError("Dataset {!s} has not implemented "
            "conversion to xarray".format(self))

class SingleFileDataset(Dataset):
    """Represents a dataset where all measurements are in one file.
    """

    start_date = datetime.datetime.min
    end_date = datetime.datetime.max
    srcfile = None

    # docstring in parent
    def find_granules(self, start=datetime.datetime.min,
                            end=datetime.datetime.max,
                            include_last_before=False):
        if start < self.end_date and end > self.start_date:
            yield self.srcfile

    # docstring in parent
    def find_granules_sorted(self, start=datetime.datetime.min,
                                   end=datetime.datetime.max,
                                   include_last_before=False):
        yield from self.find_granules(start, end)

    def get_times_for_granule(self, gran=None):
        return (self.start_date, self.end_date)

    def read(self, f=None, fields="all"):
        return super().read(f or self.srcfile, fields)

    def find_most_recent_granule_before(self, instant,
            **locator_args):
        if instant > self.start_date:
            yield from self.find_granules(**locator_args)
        else:
            raise GranuleLocatorError("Instant out of range: "
                "{:%Y-%m-%d %H:%M:%S}".format(instant))

class MultiFileDataset(Dataset):
    """Represents a dataset where measurements are spread over multiple
    files.

    If filenames contain timestamps, this information is used to determine
    the time for a granule or measurement.  If filenames do not contain
    timestamps, this information is obtained from the file contents.

    Attributes:

        basedir (pathlib.Path or str):
            Describes the directory under which all granules are located.
            Can be either a string or a pathlib.Path object.

        subdir (pathlib.Path or str):
            Describes the directory within basedir where granules are
            located.  May contain string formatting directives where
            particular fields are replaces, such as `year`, `month`, and
            `day`.  For example: `subdir = '{year}/{month}'`.  Sorting
            cannot be more narrow than by day.

        re (str):
            Regular expression that should match valid granule files within
            `basedir` / `subdir`.  Should use symbolic group names to capture
            relevant information when possible, such as starting time, orbit
            number, etc.  For time identification, relevant fields are
            contained in MultiFileDataset.date_info, where each field also
            exists in a version with "_end" appended.
            MultiFileDataset.refields contains all recognised fields.

            If any _end fields are found, the ending time is equal to the
            beginning time with any _end fields replaced.  If no _end
            fields are found, the `granule_duration` attribute is used to
            determine the ending time, or the file is read to get the
            ending time (hopefully the header is enough).

        granule_cache_file (pathlib.Path or str):
            If set, use this file to cache information related to
            granules.  This is used to cache granule times if those are
            not directly inferred from the filename.  Otherwise, this is
            not used.  The full path to this file shall be `basedir` /
            `granule_cache_file`.

        granule_duration (datetime.timedelta):
            If the filename contains starting times but no ending times,
            granule_duration is used to determine the ending time.  This
            should be a datetime.timedelta object.
    """
    basedir = None
    subdir = ""
    re = None
    _re = None # compiled version, do not touch
    granule_cache_file = None
    _granule_start_times = None
    granule_duration = None

    datefields = "year month day hour minute second".split()
    # likely extended later.  Note that "tod" is also a datefield,
    # interpreted as seconds since 00
    refields = ["".join(x)
        for x in itertools.product(datefields, ("", "_end"))]
    valid_field_values = {}

    # When set to a string, interpreted as a path to database of granules
    # with corresponding first lines.  This may be relevant for cases like
    # the NOAA and MetOp satelline sensors, where subsequent granules
    # contain some repetitions of lines from previous granules.
    granules_firstline_file = None
    granules_firstline_db = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._str2path("basedir", "subdir", "granule_cache_file")
        self._open_granule_file()
        if self.re is not None:
            self._re = re.compile(self.re, re.IGNORECASE)

    def _str2path(self, *args):
        for attr in args:
            if (getattr(self, attr) is not None and
                    not isinstance(getattr(self, attr), pathlib.PurePath)):
                setattr(self, attr, pathlib.Path(getattr(self, attr)))

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_granule_start_times"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._open_granule_file()

    def find_dir_for_time(self, dt):
        """Find the directory containing granules/measurements at (date)time

        For a given datetime object, find the directory that contains
        granules/measurument files for this particular time.

        Arguments:
            
            dt (datetime.datetime): Timestamp for inquiry.
                In reality, any object with `year`, `month`, and `day`
                attributes works.

        Returns:

            pathlib.Path object pointing to to relevant directory
        """
        return pathlib.Path(str(self.basedir / self.subdir).format(
            year=dt.year, month=dt.month, day=dt.day,
            doy=dt.timetuple().tm_yday))

    def get_mandatory_fields(self):
        fm = string.Formatter()
        fields = {f[1] for f in fm.parse(str(self.basedir / self.subdir))}
        if fields == {None}:
            fields = set()
        return fields - set(self.datefields)

    def verify_mandatory_fields(self, extra):
        mandatory = self.get_mandatory_fields()
        found = set(extra.keys())
        missing = set()
        if not mandatory <= found:
            missing = mandatory - found
            # I'm deleting elements so I must loop over a copy
            for elem in missing.copy():
                try:
                    extra[elem] = getattr(self, elem)
                except AttributeError:
                    pass
                else:
                    missing.remove(elem)
        if missing:
            raise GranuleLocatorError("Missing fields needed to search for "
                "{self.name:s} files.  Need: {mandatory:s}. "
                "Found: {found:s}.  Missing: {missing:s}".format(
                    self=self,
                    mandatory=", ".join(mandatory),
                    found=", ".join(found) or "(none)",
                    missing=", ".join(missing)))
        for field in found:
            if field in self.valid_field_values:
                if not extra[field] in self.valid_field_values[field]:
                    raise GranuleLocatorError("Encountered invalid {field!s} "
                        "when searching for {self.name!s} files. "
                        "Found: {value!s}. Valid: "
                        "{valid_values!s}".format(
                            field=field, self=self, extra=extra,
                            value=extra[field],
                            valid_values=self.valid_field_values[field]))

    def get_subdir_resolution(self):
        """Return the resolution for the subdir precision.

        Returns "year", "month", "day", or None (if there is no subdir).

        Based on parsing of self.subdir attribute.
        """
        fm = string.Formatter()
        fields = {f[1] for f in fm.parse(str(self.subdir))}
        if "day" in fields:
            return "day"
        if "month" in fields:
            if "doy" in fields:
                raise ValueError("Format string has both month and doy")
            return "month"
        if "year" in fields:
            if "doy" in fields:
                return "day"
            return "year"

    def get_path_format_variables(self):
        """What extra format variables are needed in find_granules?

        Depending on the dataset, `find_granules` needs zero or more extra
        formatting valiables.  For example, TOVS instruments require the
        satellite.  Required formatting arguments are determined by
        self.basedir and self.subdir.
        """
        return {x[1] for x in string.Formatter().parse(
                    str(self.basedir / self.subdir))} - {"year", "month",
                                                         "day"}

    def iterate_subdirs(self, d_start, d_end, **extra):
        """Iterate through all subdirs in dataset.

        Note that this does not check for existance of those directories.

        Yields a 2-element tuple where the first contains information on
        year(/month/day), and the second is the path.

        Arguments:

            d_start (datetime.date): Starting date.
            d_end (datetime.date): Ending date
            **extra: Any extra keyword arguments.  This will be passed on to
                format self.basedir / self.subdir, in case the standard fields
                like year, month, etc. do not provide enough information.

        Yields:

            pathlib.Path objects for each directory in the dataset
            containing files between `d_start` and `d_end`.
        """

        self.verify_mandatory_fields(extra)
        # depending on resolution, iterate by year, month, or day.
        # Resolution is determined by provided fields in self.subdir.
        d = d_start
        res = self.get_subdir_resolution()

        pst = str(self.basedir / self.subdir)
        if res == "year":
            year = d.year
            while datetime.date(year, 1, 1) <= d_end:
                yield (dict(year=year),
                    pathlib.Path(pst.format(year=year, **extra)))
                year += 1
        elif res == "month":
            year = d.year
            month = d.month
            while datetime.date(year, month, 1) <= d_end:
                yield (dict(year=year, month=month),
                    pathlib.Path(pst.format(year=year, month=month,
                                            **extra)))
                if month == 12:
                    year += 1
                    month = 1
                else:
                    month += 1
        elif res == "day":
            #while d < d_end:
            if any(x[1] == "doy" for x in string.Formatter().parse(pst)):
                while d <= d_end:
                    doy = d.timetuple().tm_yday
                    yield (dict(year=d.year, doy=doy),
                        pathlib.Path(pst.format(year=d.year, doy=doy,
                                                **extra)))
                    d += datetime.timedelta(days=1)
            else:
                while d <= d_end:
                    yield (dict(year=d.year, month=d.month, day=d.day),
                        pathlib.Path(pst.format(
                            year=d.year, month=d.month, day=d.day,
                            **extra)))
                    d = d + datetime.timedelta(days=1)
        else:
            yield ({}, pathlib.Path(pst.format(**extra)))
          
    def find_granules(self, dt_start=None, dt_end=None, 
            include_last_before=False, **extra):
        """Yield all granules/measurementfiles in period

        Accepts extra keyword arguments.  Meaning depends on actual
        dataset.  Could be something like a satellite name in the case of
        sensors occurring on multiple platforms, like HIRS.  To see what
        keyword arguments are accepted or possibly needed for a particular
        dataset, call self.get_path_format_variables()

        If keyword argument `return_time` is present and True, yield
        tuples of (start_time, path) rather than just `path`.

        The results are usually sorted by start time, but this is not
        guaranteed and depends on the filesystem.  If you need sorted
        granules, please use find_granules_sorted.

        Arguments:

            d_start (datetime.date): Starting date.
            d_end (datetime.date): Ending date
            include_last_before (bool): Include last granule starting
                before.
            **extra: Any extra keyword arguments.  This will be passed on to
                format self.basedir / self.subdir, in case the standard fields
                like year, month, etc. do not provide enough information.

        Yields:

            pathlib.Path objects for each datafile in the dataset between
            `dt_start` and `dt_end`.
        """

        if dt_start is None:
            dt_start = self.start_date

        if dt_end is None:
            dt_end = self.end_date

        return_time = extra.pop("return_time", False)

        self.verify_mandatory_fields(extra)

        d_start = (dt_start.date()
                if isinstance(dt_start, datetime.datetime) 
                else dt_start)
        d_end = (dt_end.date()
                if isinstance(dt_end, datetime.datetime) 
                else dt_end)
        found_any_dirs = False
        found_any_grans = False
        logger.debug(("Searching for {!s} granules between {!s} and {!s} "
                      ).format(self.name, dt_start, dt_end))
        before = None
        if include_last_before and dt_start > self.start_date:
            try:
                before = self.find_most_recent_granule_before(dt_start, 
                        return_time=return_time, **extra)
                yield before
            except GranuleLocatorError: # no problem
                pass

        for (timeinfo, subdir) in self.iterate_subdirs(d_start, d_end,
                                                       **extra):
            if subdir.exists() and subdir.is_dir():
                logger.debug("Searching directory {!s}".format(subdir))
                found_any_dirs = True
                for child in subdir.iterdir():
                    if before is not None and child == before[1]: # already yielded it as "last before"
                        continue
                    m = self._re.fullmatch(child.name)
                    if m is not None:
                        fit = True
                        for k in m.groupdict().keys() & extra.keys():
                            if m[k].lower() != extra[k].lower():
                                fit = False # not the right file
                                break
                        if not fit:
                            continue
                        found_any_grans = True
                        try:
                            (g_start, g_end) = self.get_times_for_granule(child,
                                **timeinfo)
                        except InvalidFileError as e:
                            logger.error(
                                "Skipping {!s}.  Problem: {}".format(
                                    child, e.args[0]))
                            continue
                        if g_end >= dt_start and g_start <= dt_end:
                            if return_time:
                                yield (g_start, child)
                            else:
                                yield child
        if not found_any_dirs:
            logger.warning("Found no directories.  Make sure {self.name:s} has "
                  "coverage in {dt_start:%Y-%m-%d %H:%M:%S} – "
                  "{dt_end:%Y-%m-%d %H:%M:%S} and that you spelt any "
                  "additional information correctly: ".format(self=self,
                  dt_start=dt_start, dt_end=dt_end) + str(extra) + " Satellite "
                  "names and other fields are case-sensitive!")
        elif not found_any_grans:
            logger.warning("Directories searched appear to contain no matching "
                  "files.  Make sure basedir, subdir, and regexp are "
                  "correct and you did not misspell any extra "
                  "information: " + str(extra))
            
    def find_granules_sorted(self, dt_start=None, dt_end=None, 
                include_last_before=False, **extra):
        """Yield all granules, sorted by times.

        For documentation, see :func:`~Dataset.find_granules`.
        """

        allgran = list(self.find_granules(dt_start, dt_end, 
                       include_last_before, **extra))

        # I've been through all granules at least once, so all should be
        # cached now; no need for additional hints when granule timeinfo
        # obtainable only with hints from subdir, which is not included in
        # the re-matching method
        if extra.get("return_time", False):
            yield from sorted(allgran)
        else:
            yield from sorted(allgran, key=self.get_times_for_granule)

    def find_most_recent_granule_before(self, instant, **locator_args):
        # docstring in parent class
        return_time = locator_args.pop("return_time", False)
        res = self.get_subdir_resolution()
        if res == "day":
            d0 = datetime.datetime(instant.year, instant.month, instant.day,
                                   0, 0, 0)
            d1 = d0 + datetime.timedelta(days=2) # closed interval
            d0 = d0 - datetime.timedelta(days=1) # granule may start yesterday
        elif res == "month":
            d0 = datetime.datetime(instant.year, instant.month, 1,
                                   0, 0, 0)
            d1 = d0 + datetime.timedelta(days=31*2)
            d0 = d0 - datetime.timedelta(days=31)
        elif res == "year":
            d0 = datetime.datetime(instant.year-1, 1, 1,
                                   0, 0, 0)
            d1 = datetime.datetime(instant.year+1, 12, 31, 23, 59, 59)
        else:
            d0 = d1 = None # search entire dataset

        first = True
        for (time_st, gran) in self.find_granules_sorted(d0, d1,
                        return_time=True, **locator_args):
            if time_st > instant:
                if not first:
                    if return_time:
                        return (lasttime, lastgran)
                    else:
                        return lastgran
                break
            lasttime = time_st
            lastgran = gran
            first = False
        raise GranuleLocatorError("Can not find any granule "
            "before {:%Y-%m-%d %H:%M:%S}".format(instant))

    @staticmethod
    def _getyear(gd, s, alt):
        """Extract year info from group-dict

        Taking a group dict and a string, get an int for the year.
        The group dict should come from re.fullmatch().groupdict().  The
        second argument is typically "year" or "year_end".  If this is a
        4-digit string, it is taken as a year.  If it is a 2-digit string,
        it is taken as a 2-digit year, which is taken as 19xx for <= 68,
        and 20xx for >= 69, according to POSIX and ISO C standards.
        If there is no match, return alt.

        Arguments:
            gd (dict): Group-dict such as returned by re module
            s (str): String to match for
            int (str): Value returned in case there is no match

        Returns:
            int: year
        """

        if s in gd:
            i = int(gd[s])
            if len(gd[s]) == 2:
                return 1900 + i if i> 68 else 2000 + i
            elif len(gd[s]) == 4:
                return i
            else:
                raise ValueError("Found {:d}-digit string for the year. "
                    "Expected 2 or 4 digits.  Giving up.".format(len(gd[s])))
        else:
            return alt

    def get_info_for_granule(self, p):
        """Return dict (re.fullmatch) for granule, based on re

        Arguments:
            
            p (pathlib.Path): path to granule

        Returns:
            
            dict: dictionary with info, such as returned by
            :func:`re.fullmatch`.
        """

        if not isinstance(p, pathlib.Path):
            p = pathlib.Path(p)
        m = self._re.fullmatch(p.name)
        return m.groupdict()

    def get_times_for_granule(self, p, **kwargs):
        """For granule stored in `path`, get start and end times.

        May take hints for year, month, day, hour, minute, second, and
        their endings, according to self.datefields

        Arguments:

            p (pathlib.Path): path to granule
            **kwargs: Any more info that may be needed

        Returns:
            (datetime, datetime): Start and end time for granule

        """
        if not isinstance(p, pathlib.PurePath):
            p = pathlib.PurePath(p)
        if str(p) in self._granule_start_times.keys():
            (start, end) = self._granule_start_times[str(p)]
        else:
            gd = self.get_info_for_granule(p)
            if (any(f in gd.keys() for f in self.datefields) and
                (any(f in gd.keys() for f in {x+"_end" for x in self.datefields})
                        or self.granule_duration is not None)):
                st_date = [int(gd.get(p, kwargs.get(p, 0))) for p in self.datefields]
                td = datetime.timedelta()
                if st_date[1] == st_date[2] == 0:
                    if "doy" in gd:
                        td += datetime.timedelta(days=int(gd["doy"])-1)
                    # month and day can't be 0...
                    st_date[1] = st_date[1] or 1
                    st_date[2] = st_date[2] or 1
                # maybe it's a two-year notation
                st_date[0] = self._getyear(gd, "year", kwargs.get("year", 0))

                try:
                    start = datetime.datetime(*st_date)
                except ValueError as e:
                    raise InvalidFileError("File {!s} has invalid "
                        "starting datetime, giving up".format(p)) from v
                if "tod" in gd and start.time() == datetime.time(0):
                    td += datetime.timedelta(seconds=int(gd["tod"]))
                start += td
                if any(k.endswith("_end") for k in gd.keys()):
                    # FIXME: Does this go well at the end of
                    # year/month/day boundary?  Should really makes sure
                    # that end is the first time occurance after
                    # start_time fulfilling the provided information.
                    end_date = st_date.copy()
                    end_date[0] = self._getyear(gd, "year_end", kwargs.get("year_end", st_date[0]))
                    end_date[1:] = [int(gd.get(p+"_end",
                                               kwargs.get(p+"_end", sd_x)))
                                   for (p, sd_x) in zip(self.datefields[1:],
                                                        st_date[1:])]
                    try:
                        end = datetime.datetime(*end_date)
                    except ValueError as v:
                        raise InvalidFileError("File {!s} has invalid "
                            "ending datetime, giving up".format(p)) from v
                    if end_date < st_date: # must have crossed date boundary
                        end += datetime.timedelta(days=1)
                elif self.granule_duration is not None:
                    end = start + self.granule_duration
                else:
                    raise RuntimeError("This code should never execute")
            else:
                # implementation depends on dataset
                (start, end) = self.get_time_from_granule_contents(str(p))
                self._granule_start_times[str(p)] = (start, end)
        return (start, end)

    # not an abstract method because subclasses need to implement it /if
    # and only if starting/ending times cannot be determined from the filename
    def get_time_from_granule_contents(self, p):
        """Get datetime objects for beginning and end of granule

        If it returns None, then use same as start time.

        Arguments:

            p (pathlib.Path): Path to file

        Returns:

            (datetime, datetime): 2-tuple for start and end times
        """
        raise ValueError(
            ("To determine starting and end-times for a {0} dataset, "
             "I need to read the file.  However, {0} has not implemented the "
             "get_time_from_granule_contents method.".format(
                type(self).__name__)))

    def _open_granule_file(self):
        if self.granule_cache_file is not None:
            p = str(self.basedir / self.granule_cache_file)
            try:
                self._granule_start_times = shelve.open(p, protocol=4)
            except OSError:
                logger.error(("Unable to open granule file {} RW.  "
                               "Opening copy instead.").format(p))
                tf = tempfile.NamedTemporaryFile()
                shutil.copyfile(p, tf.name)
                #self._granule_start_times = shelve.open(p, flag='r')
                self._granule_start_times = shelve.open(tf.name)
        else:
            self._granule_start_times = {}

    def _filter_firstline(self, f, M):
        # by default, do nothing
        raise NotImplementedError("Dataset {self:s} {self.name:s} needs "
            "database for storing a table with the first new scanline per "
            "granule, i.e. the first line not contained in the previous. "
            "Please define `granules_firstline` in source-code or "
            "configuration file.")

class SingleMeasurementPerFileDataset(MultiFileDataset):
    """Represents datasets where each file contains one measurement.

    An example of this would be ACE-FTS, or some radio-occultation datasets.

    Attributes:

        filename_fields (Mapping[str, dtype])

            dict with {name, dtype} for fields that should be copied from
            the filename (as obtained with self.re) into the header
    """

    granule_duration = datetime.timedelta(0)
    filename_fields = {}


    @abc.abstractmethod
    def read_single(self, p, fields="all"):
        """Read a single measurement from a single file.

        Shall take one argument (a path object) and return a tuple with
        (header, measurement).  The header shall contain information like
        latitude, longitude, time.

        Arguments:

            p (pathlib.Path): path to file
            
            fields (Iterable[str] or str): What fields to return.
                See :func:`Dataset.read_period` for details.
        """
        raise NotImplementedError()

    _head_dtype = {}
    def _read(self, p, fields="all"):
        """Reads a single measurement converted to ndarray

        Arguments:

            p (pathlib.Path): path to file

            fields (Iterable[str] or str): What fields to return.
                See :func:`Dataset.read_period` for details.
        """

        (head, body) = self.read_single(p, fields=fields)

        dt = [(s+body.shape if len(s)==2 else (s[0], s[1], s[2]+body.shape))
                for s in body.dtype.descr]
        dt.extend([("lat", "f8"), ("lon", "f8"), ("time", "M8[s]")])
        dt.extend([(s, self._head_dtype[s])
                for s in (head.keys() & self._head_dtype.keys())
                if s not in {"lat", "lon", "time"}])
        if self.filename_fields:
            info = self.get_info_for_granule(p)
            dt.extend(self.filename_fields.items())
        D = numpy.ma.empty(1, dt)

        for nm in body.dtype.names:
            D[nm] = body[nm]

        for nm in {"lat", "lon", "time"}:
            D[nm] = head[nm]

        for nm in head.keys() & D.dtype.names:
            if nm not in {"lat", "lon", "time"}:
                D[nm] = head[nm]

        if self.filename_fields:
            for nm in self.filename_fields.keys():
                D[nm] = info[nm]

        return (D, {})

class NetCDFDataset:
    """Mixin for any dataset where the contents are in NetCDF.

    This may provide a good default for any NetCDF-based dataset.  The
    reading routine will take the most commonly occurring dimension as the
    ndarray axes, and the rest within structured multidimensional dtype.

    USE WITH CARE!  PROVISIONAL API!

    This one should use xarray, of course!  See
    https://arts.mi.uni-hamburg.de/trac/rt/ticket/145
    """

    read_returns = "ndarray" # can be set to xarray

    def _get_dtype_from_vars(self, alldims, allvars, fields, prim):
        """Get dtype from alldims, allvars
        """
        return numpy.dtype([
            (k,
             "f4" if (getattr(v, "Scale", 1)!=1) else v.dtype,
             tuple(s for (i, s) in enumerate(v.shape)
                 if v.dimensions[i] != alldims[prim].name))
             for (k, v) in allvars.items()
             if ((alldims[prim].name in v.dimensions) if fields=="all"
                  else (k in fields))])

    def _read(self, f, fields="all",
              pseudo_fields=None,
              prim=None):
        if self.read_returns == "ndarray":
            return (self._read_ndarray(f, fields, pseudo_fields, prim), {})
        elif self.read_returns == "xarray":
            ds = xarray.open_dataset(f)
            if fields != "all":
                ds = ds[fields]
            if pseudo_fields is not None:
                warnings.warn("Ignoring pseudo-fields", UserWarning)
            # if there is any problem loading the file, I want to know
            # that now, not whenever I happen to read the data
            ds.load()
            return (ds, {})

    def _read_ndarray(self, f, fields="all", pseudo_fields=None,
                      prim=None):
        if pseudo_fields is None:
            pseudo_fields = {}
        with netCDF4.Dataset(f, 'r') as ds:

            # generic conversion of NetCDF to ndarray; consider the most
            # common dimension, make this one the shape of the ndarray,
            # copy over all variables that share this dimension, and
            # ignore the rest, unless variables are passed explicitly

            # first do the “pseudo fields” so I know the dtype before I
            # create the structured dtype
            extra = {nm: f(ds) for (nm, f) in pseudo_fields.items()}

            #
            # if contained in one layer of groups, flatten those first

            if ds.groups != {}: # NB: empty OrderedDict == {}
                alldims = collections.OrderedDict(
                    itertools.chain.from_iterable(
                        ds[x].dimensions.items() for x in ds.groups.keys()))
                allvars = collections.OrderedDict(
                    itertools.chain.from_iterable(
                        ds[x].variables.items() for x in ds.groups.keys()))
            else:
                alldims = ds.dimensions
                allvars = ds.variables
            # count most frequent dimensions
            cnt = collections.Counter(
                itertools.chain.from_iterable(
                    allvars[var].dimensions for var in allvars.keys()))
            prim = prim or cnt.most_common(1)[0][0]
            n = alldims[prim].size
            M = numpy.zeros(shape=(n,),
                dtype=self._get_dtype_from_vars(alldims, allvars,
                                                fields, prim).descr +
                      [(nm, v.dtype, v.shape[1:]) for (nm, v) in
                                  extra.items()])
            M = numpy.ma.masked_array(M)

            for v in M.dtype.names:
                try:
                    M[v][...] = extra[v][...]
                except KeyError:
                    # should be direct instead
                    pass
                else:
                    continue
                try:
                    M[v][...] = allvars[v][...] * getattr(allvars[v], "Scale", 1)
                    try:
                        M[v].mask[M[v]==allvars[v]._FillValue] = True
                    except AttributeError:
                        # no fill value or not a masked array...?
                        pass
                except TypeError:
                    pass # probably not a numeric type

        return M

class HomemadeDataset(NetCDFDataset, MultiFileDataset):
    """For any dataset created by typhon or its dependencies.

    Currently supports only saving to npz, through the save_npz method.
    Eventually, should also support other file formats, in particular
    NetCDF.
    """

    stored_name = None
    write_basedir = write_subdir = None
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._str2path("write_basedir", "write_subdir", "stored_name")

    def find_granule_for_time(self, **kwargs):
        """Find granule for specific time.

        May or may not exist.

        Arguments (kw only) are passed on to format directories stored in
        self.basedir / self.subdir / self.stored_name, along with
        self.__dict__.  One special keyword, "write", can be used when
        writing data, as some datasets in "production mode" have different
        files for writing than reading.

        Returns path to granule.
        """

        if kwargs.get("mode") == "write":
            d = (getattr(self, "write_basedir", self.basedir) /
                 getattr(self, "write_subdir", self.subdir) /
                 self.stored_name)
        else:
            d = self.basedir / self.subdir / self.stored_name
        # next line inspired by http://stackoverflow.com/a/32196401/974555
        # which was posted by Alek Kowalczyk on 2015-08-25 and licensed
        # under CC-BY-SA 3.0.  I believe my code is sufficiently far from
        # the original to not be a derivative work and that it can legally
        # be MIT-licensed.
        subsdict = {attr: getattr(self, attr) 
                    for attr in dir(self)}
        subsdict.update(**kwargs)
        nm = pathlib.Path(str(d).format(**subsdict))
        return nm

    def _read(self, f, fields="all"):
        if f.endswith("npz") or f.endswith("npy"):
            if fields != "all":
                warnings.warn("Ignoring fields≠'all'!", UserWarning)
            return (numpy.load(f)["arr_0"], {})
        elif f.endswith("nc"):
            try:
                return super()._read(f, fields=fields)
            except ValueError as v:
                if v.args[0] == ("zero-size array to reduction operation "
                                 "minimum which has no identity"):
                    # A bug in xarray will cause a ValueError if trying to
                    # decode the times in a NetCDF file with length 0.
                    # See https://github.com/pydata/xarray/issues/1329 .
                    # Although I can't presently solve this (not decoding
                    # would be undesirable due to inconsistencies), I can
                    # at least replace it by an exception that means "bad
                    # data".
                    raise InvalidDataError("It looks like the NetCDF file "
                        "is empty or at least has a datetime variable with "
                        "length 0.  Due to "
                        "https://github.com/pydata/xarray/issues/1329 "
                        "I can't currently handle that.  Original message "
                        "was: " + v.args[0]) from v
                else:
                    raise

#    def find_granules(self, start, end):
#        raise StopIteration()

#    @abc.abstractmethod
#    def quicksave(self, f):
#        """Quick save to file
#
#        :param str f: File to save to
#        """

    def save_npz(self, path, M):
        """Save to compressed npz

        Arguments:

            path (pathlib.Path): Path to store to

            M (ndarray): Contents of what to store.
        """
        p = pathlib.Path(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        numpy.savez_compressed(str(path), M)

class MultiSatelliteDataset(metaclass=abc.ABCMeta):
    satellites = set()
    @property
    def valid_field_values(self):
        return {"satname": self.satellites}

class HyperSpectral(Dataset, em.FwmuMixin):
    """Superclass for any hyperspectral instrument
    """
    freqfile = None

# Not yet transferred from pyatmlab to typhon, not clean enough:
#
# - ProfileDataset
# - StationaryDataset
# - HyperSpectral

class DatasetDeque:
    """A deque-like object sliding through a dataset.

    For a particular dataset, keep data corresponding to time period in
    memory.  When reading new data, discard a corresponding time period of
    data in the past.  This should be useful to process longer periods of
    data that do not fit into memory, to calculate sliding window
    statistics, or to perform processing where values nearby in time are
    required.
    """

    dsobj = window = init_time = center_time = data = None
    def __init__(self, ds, window, init_time, *args, **kwargs):
        """Initialise a DatasetDeque

        Arguments:

            ds [Dataset]: Dataset that this belongs to, for example,
                HIRS().  This dataset must have an implementation of
                .as_xarray_dataset().

            window [timedelta]: Duration that should be kept in memory.
                For example, datetime.timedelta(hours=48).

            init_time [datetime]: Instant around which initial window
                shall be centred.

            Remaining arguments passed to read_period.
        """
        self.dsobj = ds
        self.window = window
        self.init_time = init_time
        self.center_time = init_time

        self.reset(newtime=self.init_time, *args, **kwargs)

    def reset(self, newtime=None, *args, **kwargs):
        """Reset to initial conditions

        Sets data to indicated time or self.init_time ± self.window/2

        Arguments:

            newtime [datetime.datetime]: optional, time at which to
            center window.  If not given, reset to initial time.

            Remaining arguments passed to read_period.

        """
        self.edges = (self.init_time-self.window/2,
                      self.init_time+self.window/2)
        self.center_time = newtime or self.init_time
        M = self.dsobj.read_period(
            *self.edges, *args, **kwargs)

        self.data = self.dsobj.as_xarray_dataset(M)
            
    def move(self, period, *args, **kwargs):
        """Read period on the right, discard period on the left

        This moves the window to the right by `period` by reading `period`
        new data, appending this on the right, and discarding an equally
        long period on the left.

        Arguments:

            period [timedelta]: Duration by which to shift.
                Must be positive.

            Remaining arguments passed to read_period.
        """

        if period > self.window:
            raise ValueError("Shifting period ({!s}) exceeds window "
                "length ({!s})!".format(period, self.window))
        self.center_time += period
        ok = False
        try:
            Mnew = self.dsobj.read_period(
                self.edges[1],
                self.edges[1]+period,
                *args, **kwargs)
        except: # syntactically required
            raise
        else:
            datanew = self.dsobj.as_xarray_dataset(Mnew)
            ok = True
        finally:
            self.edges = (self.edges[0] + period,
                          self.edges[1] + period)

        # need to convert to ms or it will become int and indexing will
        # fail, see https://github.com/numpy/numpy/issues/8546
        # and https://github.com/pydata/xarray/issues/1240
        #
        # If the very first iteration fails to read the first period and
        # is off by X hours, EVERYTHING will be off by X hours if I simply
        # reset it based on the first time in self.data!  Therefore the
        # 'min' function.
#        newst = self.data["time"][0].values.astype("M8[ms]") + numpy.timedelta64(period)
        newst = min(numpy.datetime64(self.edges[0]), self.data["time"][0].values.astype("M8[ms]") + numpy.timedelta64(period))
        # BUG: need workaround for https://github.com/pydata/xarray/issues/1297
        all_encodings = {k: v.encoding for (k, v) in self.data.data_vars.items()}
        self.data = xarray.concat(
            (self.data.sel(time=slice(newst, None)),
             datanew),
            dim="time")
        for (k, v) in self.data.data_vars.items():
            v.encoding = all_encodings[k]

    def resize(self, window):
        """Resize window
        """

        self.window = window
        self.edges = (self.center_time-window/2, self.center_time+window/2)
        self.data = self.dsobj.as_xarray_dataset(
            self.dsobj.read_period(*self.edges))

    def __repr__(self):
        return "<{:s} {:s} centred at {:%Y-%m-%d %H:%M}>".format(
            self.__class__.__name__, repr(self.dsobj),
            self.center_time)



# -*- coding: utf-8 -*-

"""Functions for handling geographical coordinate systems
and reference ellipsoids.

Unless otherwise stated functions are ported from atmlab-2-3-181.

"""
import numpy as np

from typhon import constants

__all__ = [
    'ellipsoidmodels',
    'ellipsoid_r_geodetic',
    'ellipsoid_r_geocentric',
    'ellipsoid2d',
    'ellipsoidcurvradius',
    'sind',
    'cosd',
    'tand',
    'asind',
    'inrange',
    'cart2geocentric',
    'geocentric2cart',
    'cart2geodetic',
    'geodetic2cart',
    'geodetic2geocentric',
    'geocentric2geodetic',
    'great_circle_distance',
    'geographic_mean',
    'cartposlos2geocentric',
    'geocentricposlos2cart',
    'sphere_plane_intersection',
    'tunnel_distance',
]


def sind(x):
    """Sine of argument in degrees."""
    return np.sin(np.deg2rad(x))


def cosd(x):
    """Cosine of argument in degrees."""
    return np.cos(np.deg2rad(x))


def tand(x):
    """Tangent of argument in degrees."""
    return np.tan(np.deg2rad(x))


def asind(x):
    """Inverse sine in degrees."""
    return np.arcsin(np.deg2rad(x))


def inrange(x, minx, maxx, exclude='none', text=None):
    """Test if x is within given bounds.

    Parameters:
        x: Variable to test.
        minx: Lower boundary.
        maxx: Upper boundary.
        exclude (str): Exclude boundaries. Possible values are:
            'none', 'lower', 'upper' and 'both'
        text (str): Addiitional warning text.

    Raises:
        Exception: If value is out of bounds.

    """
    compare = {'none': (np.greater_equal, np.less_equal),
               'lower': (np.greater, np.less_equal),
               'upper': (np.greater_equal, np.less),
               'both': (np.greater, np.less),
               }

    greater, less = compare[exclude]

    if less(x, minx) or greater(x, maxx):
        if text is None:
            raise Exception('Range out of bound [{}, {}]'.format(minx, maxx))
        else:
            raise Exception(
                'Range out of bound [{}, {}]: {}'.format(minx, maxx, text)
                )


class ellipsoidmodels():
    """Provide data for different reference ellipsoids.

    The following models are covered:

        * SphericalEarth     (radius set as constants.earth_radius)
        * WGS84
        * SphericalVenus     (radius same as used in ARTS)
        * SphericalMars      (radius same as used in ARTS)
        * EllipsoidMars
        * SphericalJupiter   (radius same as used in ARTS)

    Examples:
        >>> e = ellipsoidmodels()

    .. Ported from atmlab. Original author: Patrick Eriksson
    """

    def __init__(self):
        self._data = {
            "SphericalEarth": (constants.earth_radius, 0),
            "WGS84": (6378137, 0.0818191908426),
            "SphericalVenus": (6051800.0, 0),
            "SphericalMars": (3389500.0, 0),
            "EllipsoidMars": (3396190.0, 0.1083),
            "SphericalJupiter": (69911000.0, 0),
            }

    def __getitem__(self, model):
        return self.get(model)

    def get(self, model='WGS84'):
        """Return data for different reference ellipsoids.

        Parameters:
            model (str): Model ellipsoid.

        Returns:
            tuple: Equatorial radius (r), eccentricity (e)

        Examples:
            >>> e['WGS84']
            (6378137, 0.0818191908426)
            >>> e.get('WGS84')
            (6378137, 0.0818191908426)
            >>> ellipsoidmodel()['WGS84']
            (6378137, 0.0818191908426)

        """
        if model in self._data:
            return self._data.__getitem__(model)
        else:
            raise Exception('Unknown ellipsoid model "{}".'.format(model))

    @property
    def models(self):
        """List of available models.

        Examples:
            >>> e.models
            ['SphericalVenus',
             'SphericalMars',
             'WGS84',
             'SphericalJupiter',
             'EllipsoidMars',
             'SphericalEarth']
        """
        return list(self._data.keys())


def ellipsoid_r_geocentric(ellipsoid, lat):
    """Geocentric radius of a reference ellipsoid.

    Gives the distance from the Earth's centre and the reference ellipsoid
    as a function of geo\ **centric** latitude.

    Note:
        To obtain the radii for **geodetic** latitude,
        use :func:`ellipsoid_r_geodetic`.

    Parameters:
        ellipsoid (tuple):  Model ellipsoid as returned
            by :class:`ellipsoidmodels`.
        lat (float or ndarray): Geocentric latitudes.

    Returns:
        float or ndarray: Radii.

    .. Ported from atmlab. Original author: Patrick Eriksson
    """
    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    if ellipsoid[1] == 0:
        r = np.ones(np.shape(lat)) * ellipsoid[0]
    else:
        c = 1 - ellipsoid[1]**2
        b = ellipsoid[0] * np.sqrt(c)
        r = b / np.sqrt(c * cosd(lat)**2 + sind(lat)**2)

    return r


def ellipsoid_r_geodetic(ellipsoid, lat):
    """Geodetic radius of a reference ellipsoid.

    Gives the distance from the Earth's centre and the reference ellipsoid
    as a function of geo\ **detic** latitude.

    Note:
        To obtain the radii for **geocentric** latitude,
        use :func:`ellipsoid_r_geocentric`.

    Parameters:
        ellipsoid (tuple):  Model ellipsoid as returned
            by :class:`ellipsoidmodels`.
        lat (float or ndarray): Geodetic latitudes.

    Returns:
        float or ndarray: Radii.

    .. Ported from atmlab. Original author: Patrick Eriksson
    """
    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    if ellipsoid[1] == 0:
        r = np.ones(np.shape(lat)) * ellipsoid[0]
    else:
        e2 = ellipsoid[1]**2
        sin2 = sind(lat)**2
        r = (ellipsoid[0] * np.sqrt((1 - e2)**2 * sin2 +
             cosd(lat) ** 2) / np.sqrt(1 - e2 * sin2))
    return r


def ellipsoid2d(ellipsoid, orbitinc):
    """Approximate ellipsoid for 2D calculations.

    Determines an approximate reference ellipsoid following an orbit track.
    The new ellipsoid is determined simply, by determining the radius at the
    maximum latitude and from this value calculate a new eccentricity.
    The orbit is specified by giving the orbit inclination, that is
    normally a value around 100 deg for polar sun-synchronous orbits.

    Parameters:
        ellipsoid (tuple):  Model ellipsoid as returned
            by :class:`ellipsoidmodels`.
        orbitinc (float): Orbit inclination.

    Returns:
        tuple: Modified ellipsoid vector.

    .. Ported from atmlab. Original author: Patrick Eriksson
    """
    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    inrange(orbitinc, 0, 180,
            exclude='both',
            text='Invalid orbit inclination.')

    rp = ellipsoid_r_geocentric(ellipsoid, orbitinc)

    return ellipsoid[0], np.sqrt(1 - (rp / ellipsoid[0])**2)


def ellipsoidcurvradius(ellipsoid, lat_gd, azimuth):
    """Sets ellispoid to local curvature radius

    Calculates the curvature radius for the given latitude and azimuth
    angle, and uses this to set a spherical reference ellipsoid
    suitable for 1D calculations. The curvature radius is a better
    local approximation than using the local ellipsoid radius.

    For exact result the *geodetic* latitude shall be used.

    Parameters:
        lat_gd: Geodetic latitude.
        azimuth: Azimuthal angle (angle from NS plane).
            If given curvature radii are returned, see above.

    Returns:
        tuple: Modified ellipsoid.

    .. Ported from atmlab. Original author: Patrick Eriksson
    """
    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    aterm = 1 - ellipsoid[1]**2 * sind(lat_gd)**2
    rn = 1 / np.sqrt(aterm)
    rm = (1 - ellipsoid[1]**2) * (rn / aterm)
    e0 = (ellipsoid[0] / (cosd(azimuth)**2.0 / rm + sind(azimuth)**2.0 / rn))
    e1 = 0

    return e0, e1


def cart2geocentric(x, y, z, lat0=None, lon0=None, za0=None, aa0=None):
    """Convert cartesian position to spherical coordinates.

    The geocentric Cartesian coordinate system is fixed with respect to the
    Earth, with its origin at the center of the ellipsoid and its X-, Y-,
    and Z-axes intersecting the surface at the following points:

        * X-axis: Equator at the Prime Meridian (0°, 0°)

        * Y-axis: Equator at 90-degrees East (0°, 90°)

        * Z-axis: North Pole (90°, 0°)

    A common synonym is Earth-Centered, Earth-Fixed coordinates, or ECEF.

    If the optional arguments are given, it is ensured that latitude and
    longitude are kept constant for zenith or nadir cases, and the longitude
    for N-S cases. The optional input shall be interpreted as the [x,y,z]
    is obtained by moving from [lat0,lon0] in the direction of [za0,aa0].

    Parameters:
        x: Coordinate in x dimension.
        y: Coordinate in y dimension.
        z: Coordinate in z dimension.
        lat0: Original latitude.
        lon0: Original longitude.
        za0: Orignal zenith angle.
        aa0: Orignal azimuth angle.

    Returns:
        tuple: Radius, Latitude, Longitude

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    r = np.sqrt(x**2 + y**2 + z**2)

    if np.any(r == 0):
        raise Exception("This set of functions does not handle r = 0.")

    lat = np.rad2deg(np.arcsin(z / r))
    lon = np.rad2deg(np.arctan2(y, x))

    if all(x is not None for x in [lat0, lon0, za0, aa0]):
        for i in range(np.size(r)):
            if za0[i] < 1e-06 or za0[i] > 180 - 1e-06:
                lat[i] = lat0[i]
                lon[i] = lon0[i]

            if (abs(lat0[i]) < 90 - 1e-08 and
               (abs(aa0[i]) < 1e-06 or abs(aa0[i] - 180) < 1e-06)):
                if abs(lon[i] - lon0[i]) < 1:
                    lon[i] = lon0[i]
                else:
                    if lon0[i] > 0:
                        lon[i] = lon0[i] - 180
                    else:
                        lon[i] = lon0[i] + 180

    return r, lat, lon


def geocentric2cart(r, lat, lon):
    """Convert from spherical coordinate to a cartesian position.

     See :func:`cart2geocentric` for a defintion of the geocentric
     coordinate system.

     Parameters:
            r: Radius.
            lat: Latitude in degree.
            lon  Longitude in degree.

     Returns:
        tuple: Coordinate in x, y, z dimension.

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    if np.any(r == 0):
        raise Exception("This set of functions does not handle r = 0.")

    latrad = np.deg2rad(lat)
    lonrad = np.deg2rad(lon)

    x = r * np.cos(latrad)
    y = x * np.sin(lonrad)
    x = x * np.cos(lonrad)
    z = r * np.sin(latrad)

    return x, y, z


def cart2geodetic(x, y, z, ellipsoid=None):
    """Convert from cartesian to geodetic coordinates.

    The geodetic coordinates refer to the reference ellipsoid
    specified by input ellipsoid.
    See module docstring for a defintion of the geocentric coordinate system.

    Parameters:
        x: Coordinates in x dimension.
        y: Coordinates in y dimension.
        z: Coordinates in z dimension.
        ellipsoid: A tuple with the form (semimajor axis, eccentricity).
            Default is 'WGS84' from :class:`ellipsoidmodels`.

    Returns:
        tuple: Geodetic height, latitude and longitude

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    if ellipsoid is None:
        ellipsoid = ellipsoidmodels()['WGS84']

    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    lon = np.rad2deg(np.arctan2(y, x))
    B0 = np.arctan2(z, np.hypot(x, y))
    B = np.ones(B0.shape)
    e2 = ellipsoid[1]**2
    if e2 == 0.0:
        h, lat, lon = cart2geocentric(x, y, z)
        h -= ellipsoid[0]
    else:
        while (np.any(np.abs(B - B0) > 1e-10)):
            N = ellipsoid[0] / np.sqrt(1 - e2 * np.sin(B0)**2)
            h = np.hypot(x, y) / np.cos(B0) - N
            B = B0.copy()
            B0 = np.arctan(z/np.hypot(x, y) * ((1-e2*N/(N+h))**(-1)))

        lat = np.rad2deg(B)

    return h, lat, lon


def geodetic2cart(h, lat, lon, ellipsoid=None):
    """Convert from geodetic to geocentric cartesian coordinates.

    The geodetic coordinates refer to the reference ellipsoid
    specified by input ellipsoid.
    See module docstring for a defintion of the geocentric coordinate system.

    Parameters:
        h: Geodetic height (height above the reference ellipsoid).
        lat: Geodetic latitude.
        lon: Geodetic longitude.
        ellipsoid: A tuple with the form (semimajor axis, eccentricity).
            Default is 'WGS84' from :class:`ellipsoidmodels`.

    Returns:
        tuple: x, y, z coordinates.

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    if ellipsoid is None:
        ellipsoid = ellipsoidmodels()['WGS84']

    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    a = ellipsoid[0]
    e2 = ellipsoid[1] ** 2

    N = a / np.sqrt(1 - e2 * sind(lat)**2)
    x = (N + h) * (cosd(lat)) * (cosd(lon))
    y = (N + h) * (cosd(lat)) * (sind(lon))
    # np.ones(np.shape(lon)): Ensure equal shape of x, y, z.
    z = (N * (1 - e2) + h) * (sind(lat)) * np.ones(np.shape(lon))

    return x, y, z


def geodetic2geocentric(h, lat, lon, ellipsoid=None, **kwargs):
    """Convert from geodetic to geocentric coordinates.

    The geodetic coordinates refer to the reference ellipsoid
    specified by input ellipsoid.
    See module docstring for a defintion of the geocentric coordinate system.

    Parameters:
        h: Geodetic height (height above the reference ellipsoid).
        lat: Geodetic latitude.
        lon: Geodetic longitude.
        kwargs: Additional keyword arguments for :func:`cart2geocentric`.
        ellipsoid: A tuple with the form (semimajor axis, eccentricity).
            Default is 'WGS84' from :class:`ellipsoidmodels`.

    Returns:
        tuple: Radius, geocentric latiude, geocentric longitude

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    if ellipsoid is None:
        ellipsoid = ellipsoidmodels()['WGS84']

    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    cart = geodetic2cart(h, lat, lon, ellipsoid)
    return cart2geocentric(*cart, **kwargs)


def geocentric2geodetic(r, lat, lon, ellipsoid=None):
    """Convert from geocentric to geodetic coordinates.

    The geodetic coordinates refer to the reference ellipsoid
    specified by input ellipsoid.
    See module docstring for a defintion of the geocentric coordinate system.

    Returns:
        tuple: Geodetic height, latitude and longitude

    Parameters:
        r: Radius:
        lat: Geocentric latitude.
        lon: Geocentric longitude.
        ellipsoid: A tuple with the form (semimajor axis, eccentricity).
            Default is 'WGS84' from :class:`ellipsoidmodels`.

    .. Ported from atmlab. Original author: Bengt Rydberg
    """
    if ellipsoid is None:
        ellipsoid = ellipsoidmodels()['WGS84']

    errtext = 'Invalid excentricity value in ellipsoid model.'
    inrange(ellipsoid[1], 0, 1, exclude='upper', text=errtext)

    cart = geocentric2cart(r, lat, lon)
    return cart2geodetic(*cart, ellipsoid)


def great_circle_distance(lat1, lon1, lat2, lon2, r=None):
    r"""Calculate the distance between two geographical positions

    This is the 'As-the-crow-flies' distance between two points, specified by
    their latitude and longitude. This uses the so-called haversine formula,
    defined by:

    .. math::
        d = 2 \arcsin \left( \sqrt{
            \sin \left(\frac{\varphi_2-\varphi_1}{2} \right)^2
            + \cos(\varphi_1) \cdot \cos(\varphi_1)
            \cdot \sin \left(\frac{\lambda_2-\lambda_1}{2} \right)^2} \right)

    Args:
        lat1: Latitude of position 1.
        lon1: Longitude of position 1.
        lat2: Latitude of position 2.
        lon2: Longitude of position 2.
        r (float): The radius (common for both points).

    Returns:
        If the optional argument *r* is given, the distance in m is returned.
        Otherwise the angular distance in degrees is returned.

    .. Taken from https://stackoverflow.com/a/29546836/9144990
    """
    lon1, lat1, lon2, lat2 = map(np.radians, [lon1, lat1, lon2, lat2])

    dlon = lon2 - lon1
    dlat = lat2 - lat1

    a = np.sin(dlat / 2.0) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(
        dlon / 2.0) ** 2

    c = 2 * np.arcsin(np.sqrt(a))

    if r is None:
        return np.rad2deg(c)
    else:
        return r * c


def geographic_mean(lat, lon, h=0, ellipsoid=None):
    """Calculate mean position for set of coordinates.

    Parameters:
        lat (float or ndarray): Latitudes in degrees.
        lon (float or ndarray): Longitudes in degrees.
        h (float or ndarray): Optiional altitude for
            each coordinate (default is).
        ellipsoid: A tuple with the form (semimajor axis, eccentricity).
            Default is 'WGS84' from :class:`ellipsoidmodels`.

    Returns:
       tuple: Mean latitudes and longitudes in degrees.

    """
    if ellipsoid is None:
        ellipsoid = ellipsoidmodels()['WGS84']

    x, y, z = geodetic2cart(
        h,
        lat,
        lon,
        ellipsoid=ellipsoid)

    mh, mlat, mlon = cart2geodetic(
        np.mean(x),
        np.mean(y),
        np.mean(z),
        ellipsoid=ellipsoid)

    return mlat, mlon


def cartposlos2geocentric(x, y, z, dx, dy, dz, ppc=None,
                          lat0=None, lon0=None, za0=None, aa0=None):
    """Convert cartesian POS/LOS to spherical coordinates.

    Position is given as (x,y,z), while line-of-sight is given as (dx,dy,dz).
    The corresponding quantities in polar coordinates are (r,lat,lon)
    and (za,aa), respectively.

    See *Contents* for defintion of coordinate systems.

    If the optional arguments are given, it is ensured that latitude and
    longitude are kept constant for zenith or nadir cases, and the longitude
    and azimuth angle for N-S cases. The optional input shall be interpreted
    as the [x,y,z] is obtained by moving from [r0,lat0,lon0] in the direction
    of [za0,aa0].

    This version is different from the atmlab version by normalizing the los-
    vector and demanding all or nothing for the optional arguments to work.

    Parameters:
        x: Coordinate in x dimension.
        y: Coordinate in y dimension.
        z: Coordinate in z dimension.
        dx: LOS component in x dimension.
        dy: LOS component in y dimension.
        dz: LOS component in z dimension.
        ppc: Propagation path constant = r0*sin(za0).
        lat0: Original latitude.
        lon0: Original longitude.
        za0: Orignal zenith angle.
        aa0: Orignal azimuth angle.

    Returns:
        tuple(ndarray): Radius, Latitude, Longitude,
            Zenith angle, Azimuth angle

    .. Ported from atmlab. Original author: Bengt Rydberg

    """
    # Here be dragons!

    # Broadcast all input variables to the same shape.  Atleast (1)
    if(ppc is not None and za0 is not None and lat0 is not None and
       aa0 is not None and lon0 is not None):
        x, y, z, dx, dy, dz, ppc, lat0, lon0, za0, aa0 = _broadcast(
            x, y, z, dx, dy, dz, ppc, lat0, lon0, za0, aa0)
    elif ppc is not None:
        x, y, z, dx, dy, dz, ppc = _broadcast(x, y, z, dx, dy, dz, ppc)
    else:
        x, y, z, dx, dy, dz = _broadcast(x, y, z, dx, dy, dz)

    r, lat, lon = cart2geocentric(x, y, z, lat0, lon0, za0, aa0)

    # Renormalize for length of the variables (not in atmlab)
    norm_r = np.sqrt(dx**2 + dy**2 + dz**2)
    dx = dx / norm_r
    dy = dy / norm_r
    dz = dz / norm_r

    coslat = np.cos(np.deg2rad(lat))
    sinlat = np.sin(np.deg2rad(lat))
    coslon = np.cos(np.deg2rad(lon))
    sinlon = np.sin(np.deg2rad(lon))
    dr = np.clip(coslat * coslon * dx + sinlat * dz + coslat * sinlon * dy,
                 -1., 1.)

    # Get LOS angle
    if ppc is None:
        za = np.rad2deg(np.arccos(dr))
    else:
        za = np.rad2deg(np.arcsin(ppc / r))
    aa = np.zeros(za.shape)

    # Fix zenith and azimuth angle with optional input only when all exists
    if(za0 is not None and lat0 is not None and
       aa0 is not None and lon0 is not None):

        # Determine the type for zenith
        noz = np.logical_or(za0 < 1e-06, za0 > 180 - 1e-06)
        nan = np.isnan(za)
        pre = np.logical_and(~noz, nan)

        # Either set or do not
        za[noz] = za0[noz]
        za[pre] = 90.
        # NB: removed check for dr<0 since by putting dr==1 is more sensible

        # Determine the type for azimuth
        cir1 = abs(aa0) < 1e-06
        cir2 = np.logical_or(cir1, abs(aa0 - 180) < 1e-06)
        same = np.equal(lon, lon0)
        circ = np.logical_and(cir2, same)
        left = np.logical_and(cir1, ~same)
        right = np.logical_and(~cir1, ~same)

        # This should set all cases
        aa[circ] = aa0[circ]
        aa[left] = 180.
        aa[right] = 0.
    else:

        # Determine the type of calculations to be carried out
        noz = np.logical_or(za < 1e-06, za > 180 - 1e-06)
        pol = abs(lat) > 90 - 1e-08
        pre = np.logical_and(~noz, pol)
        non = np.logical_and(~noz, ~pol)
        aa[noz] = 0.
        aa[pre] = np.rad2deg(np.arctan2(dy[pre], dx[pre]))

        dlat = (- sinlat[non] * coslon[non] / r[non] * dx[non] + coslat[non] /
                r[non] * dz[non] - sinlat[non] * sinlon[non] / r[non] * dy[non]
                )
        dlon = (- sinlon[non] / coslat[non] / r[non] * dx[non] + coslon[non] /
                coslat[non] / r[non] * dy[non])
        aa[non] = (np.rad2deg(np.arccos(r[non] *
                   dlat / np.sin(np.deg2rad(za[non])))))

        fix = np.logical_or(np.isnan(aa), ~np.isreal(aa))

        aa[np.logical_and(fix, dlat >= 0)] = 0
        aa[np.logical_and(fix, dlat < 0)] = 180

        aa[np.logical_and(~fix, dlon < 0)] *= -1

    return r, lat, lon, za, aa


def geocentricposlos2cart(r, lat, lon, za, aa):

    """
     Convert from spherical POS/LOS to cartesian POS/LOS

     See Contents for a defintion of the geocentric coordinate system.
     The local LOS angles are defined following the EAST-NORTH-UP system:

             za    aa

             90    0   points towards north

             90    90  points towards east

             0     aa  points up

     FORMAT  [x,y,z,dx,dy,dz]=geocentricposlos2cart(r,lat,lon,za,az)

     OUT
            x    Coordinate in x dimension

            y    Coordinate in y dimension

            z    Coordinate in z dimension

            dx   LOS component in x dimension

            dy   LOS component in y dimension

            dz   LOS

     IN
            r    Radius

            lat  Latitude

            lon  Longitude

            za   zenith angle

            aa   azimuth angle

     Ported from atmlab.  Original author: Bengt Rydberg 2011-10-31
    """

    r, lat, lon, za, aa = _broadcast(r, lat, lon, za, aa)

    if any(r == 0):
        raise Exception("This function is not handling the case of r = 0.")
    if any(lat < -90) or any(lat > 90):
            raise RuntimeError("The latitude is out of range")
    if any(lon < -180) or any(lon > 180):
            raise RuntimeError("The longitude is out of range")
    if any(za < 0) or any(za > 180):
            raise RuntimeError("The zenith angle is out of range")

    deg2rad = np.deg2rad(1)

    x = np.empty(r.shape)
    y = np.empty(r.shape)
    z = np.empty(r.shape)
    dx = np.empty(r.shape)
    dy = np.empty(r.shape)
    dz = np.empty(r.shape)

    at_pole = abs(lat) > (90 - 1e-8)

    if any(at_pole):
        s = np.sign(lat[at_pole])
        x[at_pole] = 0.
        y[at_pole] = 0.
        z[at_pole] = s * r[at_pole]
        dz[at_pole] = s * np.cos(deg2rad*(za[at_pole]))
        dx[at_pole] = np.sin(deg2rad*(za))
        dy[at_pole] = dx[at_pole] * np.sin(deg2rad*(aa))
        dx[at_pole] *= np.cos(deg2rad*(aa))

    not_pole = np.logical_not(at_pole)

    if any(not_pole):
        latrad = deg2rad * lat[not_pole]
        lonrad = deg2rad * lon[not_pole]
        zarad = deg2rad * za[not_pole]
        aarad = deg2rad * aa[not_pole]

        coslat = np.cos(latrad)
        sinlat = np.sin(latrad)
        coslon = np.cos(lonrad)
        sinlon = np.sin(lonrad)
        cosza = np.cos(zarad)
        sinza = np.sin(zarad)
        cosaa = np.cos(aarad)
        sinaa = np.sin(aarad)

        x[not_pole] = r[not_pole] * coslat
        y[not_pole] = x[not_pole] * sinlon
        x[not_pole] *= coslon
        z[not_pole] = r[not_pole] * sinlat

        dr = cosza
        dlat = sinza * cosaa
        dlon = sinza * sinaa / coslat

        dx[not_pole] = (coslat * coslon * dr - sinlat * coslon * dlat -
                        coslat * sinlon * dlon)
        dz[not_pole] = sinlat * dr + coslat * dlat
        dy[not_pole] = (coslat * sinlon * dr - sinlat * sinlon * dlat +
                        coslat * coslon * dlon)

    return x, y, z, dx, dy, dz


def get_ellipsoid_semiminor_axis(ellipsoid):
    """Returns the semiminor axis for the ellipsoid
    """
    return ellipsoid[0] * np.sqrt(1.0 - ellipsoid[1]**2)


def line_ellipsoid_intersect(x, y, z, dx, dy, dz,
                             ellipsoid, altitude=0.0):
    """ Finds positions of intersection of line with ellipsoid

    Solves for d:
        $$(X/a)**2 + (Y/a)**2 + (Z/b)**2 - 1 = 0$$
        $$X = x + d * dx$$
        $$Y = y + d * dy$$
        $$Z = z + d * dz$$
    where d is a multiplier of sqrt(dx**2 + dy**2 + dz**2), a is the
    semimajor axis and b is the semiminor axis. X, Y, Z gives the
    point of intersection.  If dx, dy, dz is normalized, d is distance
    in ellipsoid size units.  If no intersect, the code returns np.nan distance

    For real d, the halfway point between the points is the tangent point,
    though it is below the ellipsoid + altitude surface

    Returns:
        d: Array of distance parameters, last dimension gives the 2 solutions

    Parameters:
        x: Geocentric cartesian x-coordinate

        y: Geocentric cartesian y-coordinate

        z: Geocentric cartesian z-coordinate

        dx: Geocentric cartesian dx-view

        dy: Geocentric cartesian dy-view

        dz: Geocentric cartesian dz-view

        elliposid: ellipsoid model [a, e]

        altitude: altitude above elipsoid to intersect at [defaults to zero]

    Todo:
        Refer to line_sphere_intersect if ellipsoid is sphere
    """

    # semimajor axis
    a = ellipsoid[0] + altitude

    # semiminor axis
    b = get_ellipsoid_semiminor_axis(ellipsoid) + altitude

    # If these are scalars make them arrays
    x, y, z, dx, dy, dz = _broadcast(x, y, z, dx, dy, dz)

    # A*d**2 + B*d + C = 0, solve for d
    A = ((dx**2 + dy**2) / a**2 + dz**2 / b**2).flatten()
    B = (2 * ((x * dx + y * dy) / a**2 + z * dz / b**2)).flatten()
    C = ((y**2 + x**2) / a**2 + z**2 / b**2 - 1.0).flatten()

    d = np.zeros((len(A), 2))
    for i in range(len(A)):
        roots = np.roots([A[i], B[i], C[i]])
        if any(np.isreal(roots)):
            d[i, 0] = roots[0]
            d[i, 1] = roots[1]
        else:
            d[i, 0] = np.nan
            d[i, 1] = np.nan

    sh = []
    for i in x.shape:
        sh.append(i)
    sh.append(2)

    return d.reshape(tuple(sh))


def geometric_limb_zenith_angle(ellipsoid, r, lat, lon,
                                aa=0.0, alt=0.0, za_acc=1e-10):
    """Numerical method to compute the geometric limb zenith angle knowing
    other satellite parameters

    Assumes the planet is an ellipsoid and calculates the limb by numerically
    minimizing the distance between adjacent intersections of the ellipsoid.

    The rate of closing in on the right zenith angle should follow 180/30**N,
    though numerical errors at very exact accuracies are not guarded against.

    Parameters:

        ellipsoid: ellipsoid [a, e]

        r: geocentric radius of satellite

        lat: geocentric latitude of satellite

        lon: geocentric longitude of satellite

        aa: azimuth angle of satellite

        alt: tangent altitude

        za_acc: zenith angle accuracy.  Warning: Endless loop if too small

    Returns:

        za: satellite zenith angle for limb view geometry. Shaped as input
    """

    if(za_acc == 0):
        raise RuntimeError("Zenith accuracy cannot be 0")

    r, lat, lon, aa, alt = _broadcast(r, lat, lon, aa, alt)

    # Remember shape because we'll flatten these arrays for easy looping
    sh = r.shape
    r = r.flatten()
    lat = lat.flatten()
    lon = lon.flatten()
    aa = aa.flatten()
    alt = alt.flatten()

    za = np.empty_like(r)

    for i in range(len(r)):
        za_min = 0.0
        za_max = 180.0

        while True:  # Until zenith accuracy is reached

            # num=31 means 180/30**N is the number of loops to reach accuracy
            zenith_angles = np.linspace(za_min, za_max, num=31)

            # Get the Cartesian views for the probing zenith accuracies
            x, y, z, dx, dy, dz = geocentricposlos2cart(r[i], lat[i], lon[i],
                                                        zenith_angles, aa[i])

            # Get intersections with the ellipsoid at the tangent altitude
            d = line_ellipsoid_intersect(x, y, z, dx, dy, dz,
                                         ellipsoid, alt[i])

            # If we have somehow designed a scenario that is impossible or bad
            if np.isnan(d).all():
                raise RuntimeError("Not possible to get a tangent point.  " +
                                   "All tries miss the ellipsoid.")
            elif not np.isnan(d).any():
                raise RuntimeError("Not possible to get a tangent point.  " +
                                   "Inside ellipsoid or atmosphere.")

            # Limb zenith from minimizing distance between intersections
            D = d[:, 0] - d[:, 1]

            # These are the points of interest
            these = np.nan_to_num(D) > 0

            # The first of the zenith angles hitting the atmosphere is updated
            za_max = zenith_angles[these][0]

            # The last of the zenith angles missing the atmosphere is updated
            za_min = zenith_angles[np.logical_not(these)][-1]

            # Are we there yet?
            if (za_max-za_min) < za_acc:
                break

        # Note assignment to flat zenith angle array
        za[i] = za_max

    # Note that we return with the proper shape
    return za.reshape(sh)


def sphere_plane_intersection(pos, r, theta=np.linspace(-180., 180.)):
    """Computes the intersection of a plane with a sphere

    Returns cartesian x, y, z at the provided angles.  Sphere is centered
    at 0, 0, 0

    The solution is from

    .. math::
        \\vec{r}_c = \\vec{r}_p + r_0 \\left( \\vec{v}_1\\cos\\theta +
        \\vec{v}_2\\sin\\theta \\right)

    rp is the input position of the center of the interesecting circle,

    .. math::
        r_0 = \\sqrt{r^2 - \\vec{r_p} \\cdot \\vec{r_p}},\\;\\;
        \\vec{n} = \\vec{r}_p/||\\vec{r}_p||,\\;\\;
        \\vec{n}\\cdot\\vec{v}_1\\approx0,\\;\\;
        \\vec{v}_2 = \\vec{n} \\times \\vec{v}_1

    where v1 is so defined that it fully overlaps with the azimuth angle at the
    equator.  At other latitudes, the azimuth and theta are only the same at 0
    degrees.

    Parameters:
        pos (N, 3-array): position at the center of the circle-like shape that
        the intersection will form (i.e., the tangent point of a satellite at
        the sphere's radius).  Cannot be the 0-vector

        r (float): radius of sphere in same units as pos. Cannot be 0.0

        theta (n-array): determines positions by angle [degrees]

    Returns:
        (n, N, 3)-array of points intersecting plane and sphere

    Examples:
        Find possible positions that a satellite at an orbit about
        600 km above the surface can see a tangent point 30 km above
        the surface of Earth

        >>> x, y, z = geodetic2cart(30e3, 30., 30.)
        >>> sat = sphere_plane_intersection(np.array([x, y, z]), 6978e3)
    """
    assert r > 0.0, "Cannot work with zero radius sphere"

    assert len(pos.shape) == 2, "Need (N, 3) as pos-shape"

    d = np.linalg.norm(pos, axis=1)
    assert all(d > 0.0), "Cannot work with the 0-vector for position"

    number_of_angles = len(theta)
    number_of_positions = pos.shape[0]
    points = np.empty((number_of_angles, number_of_positions, 3))
    if number_of_angles == 0:
        return points

    # If the sphere is under the plane, or tangential to the plane
    for ii in range(number_of_positions):
        if d[ii] > r:
            for i in range(number_of_angles):
                points[i, ii, :] = np.array([np.nan, np.nan, np.nan])
            continue
        elif d[ii] == r:
            for i in range(number_of_angles):
                points[i, ii, :] = pos[ii, :]
            continue

        # The (normalized/unit) plane vector
        n = pos[ii] / d[ii]

        # Southern vector
        ones, lat, lon = cart2geocentric(n[0], n[1], n[2])

        if(lat == 90.0):
            v1 = np.array([cosd(lon), sind(lon), 0.]) * sind(lat)
        else:
            v1 = np.array([cosd(lon) * sind(lat),
                           sind(lon) * sind(lat),
                           cosd(lat-180.0)])

        # Eastern vector
        v2 = -np.cross(n, v1)  # points East

        # Radius of the circle
        rc = np.sqrt(r**2 - d[ii]**2)

        for i in range(number_of_angles):
            points[i, ii, :] = pos[ii, :] + rc * (cosd(theta[i]) * v1 +
                                                  sind(theta[i]) * v2)
    return points


def tunnel_distance(lat1, lon1, lat2, lon2):
    """Calculate the tunnel distance between two points on the Earth's surface

    Args:
        lat1: Single number or numpy array with latitudes in degrees.
        lon1: Single number or numpy array with longitudes in degrees.
        lat2: Single number or numpy array with latitudes in degrees.
        lon2: Single number or numpy array with longitudes in degrees.

    Returns:
        Tunnel distance in meters
    """
    points1 = np.column_stack(
        geocentric2cart(constants.earth_radius, lat1, lon1)
    )
    points2 = np.column_stack(
        geocentric2cart(constants.earth_radius, lat2, lon2)
    )
    return np.sqrt(np.sum((points2 - points1)**2, axis=1))


def _broadcast(*args):
    """ Similar to broadcast_arrays in numpy but with minimum output size (1,)
    """
    shape = np.broadcast(*args).shape
    if not shape:
        shape = (1,)
    return [np.broadcast_to(array, shape) for array in args]



# -*- coding: utf-8 -*-

"""This module contains custom roles to use in Sphinx.
"""
from docutils import nodes


def setup(app):
    """Install the extension.

    Parameters:
        app: Sphinx application context.
    """
    app.add_role('arts', arts_docserver_role)


def arts_docserver_role(name, rawtext, text, lineno, inliner, options=None,
                        content=None):
    """Create a link to ARTS docserver.

    Parameters:
        name (str): The role name used in the document.
        rawtext (str): The entire markup snippet, with role.
        text (str): The text marked with the role.
        lineno (str): The line number where rawtext appears in the input.
        inliner (str): The inliner instance that called us.
        options (dict): Directive options for customization.
        content (list): The directive content for customization.

    Returns:
     list, list: Nodes to insert into the document, System messages.
    """
    if content is None:
        content = []

    if options is None:
        options = {}

    url = 'http://radiativetransfer.org/docserver-trunk/all/{}'.format(text)
    node = nodes.reference(rawtext, text, refuri=url, **options)

    return [node], []



import xarray as xr

from .common import NetCDF4, expects_file_info

__all__ = [
    'HOAPS',
]


class HOAPS(NetCDF4):
    """File handler that can read data from HOAPS NetCDF4 files.

    This object handles HOAPS V3.3 NetCDF4 files such as they are
    compatible with :mod:`typhon.collocations`, i.e.:

    * convert the content of the *time* variable to numpydtetime64 objects.

    Examples:

        Draw a world map with all measurements of OceanRAIN:

        .. :code-block:: python3

            from typhon.files import FileSet, HOAPS

            hoaps = FileSet(
                name="HOAPS",
                path=".../HOAPS_v3.3/{year}-{month}/hoaps-s.f16.{year}-{month}-{day}.nc",
                handler=HOAPS(),
                # This extracts the field *asst*:
                read_args={
                    "fields": ["asst"],
                },
                # The path of HOAPS files does not provide the end of the
                # file's time coverage. Hence, we set it here explicitly:
                time_coverage="1 day",
            )

            print(hoaps["2013"])

    """

    def __init__(self, **kwargs):
        """Initializes a HOAPS file handler class

        Args:
            **kwargs: Additional key word arguments that are allowed for the
                :class:`typhon.files.handlers.common.NetCDF4` class.

        """
        # Call the base class initializer
        super().__init__(**kwargs)

    @expects_file_info()
    def get_info(self, file, **kwargs):
        """Get information about an HOAPS dataset file

        Args:
            file: A string containing path and name or a :class:`FileInfo`
                object of the file of which to get the information about.
            **kwargs: Additional keyword arguments.

        Returns:
            A :class:`FileInfo` object.
        """
        data = super().read(file, fields=("time",))

        file.times = [data["time"].min().item(0), data["time"].max().item(0)]

        return file

    @expects_file_info()
    def read(self, filename, **kwargs):
        """Read and parse a NetCDF file and load it to a xarray.Dataset

        Args:
            filename: Path and name of the file as string or FileInfo object.
            **kwargs: Additional key word arguments that are allowed for the
                :class:`~typhon.files.handlers.common.NetCDF4` class.

        Returns:
            A xarray.Dataset object.
        """

        # Make sure that the standard fields are always gonna be imported:
        fields = kwargs.pop("fields", None)
        if fields is not None:
            fields = {"time", "lat", "lon"} | set(fields)

        # xarray has problems with decoding the time variable correctly. Hence,
        # we disable it here:
        decode_cf = kwargs.pop("decode_cf", True)

        data = super().read(filename, fields=fields, decode_cf=False, **kwargs)

        # Then we fix the problem (we need integer64 instead of integer 32):
        attrs = data["time"].attrs.copy()
        data["time"] = data["time"].astype(int)
        data["time"].attrs = attrs

        # Do decoding now (just if the user wanted it!)
        if decode_cf:
            return xr.decode_cf(data)

        return data



# -*- coding: utf-8 -*-

"""Functions related to water vapor and its thermodynamic effects

This module contains wrapper functions to perform calculations with pint
quantities.

"""
from typhon import physics

from typhon.physics.units import constants
from typhon.physics.units.common import ureg


__all__ = [
    'density',
]


def density(p, T, R=None):
    """Wrapper around :func:`typhon.physics.thermodynamics.density`.

    Parameters:
        p (Quantity): Pressure.
        T (Quantity): Temperature.
            If magnitude of T and p is ndarray, sizes must match.
        R (Quantity): Gas constant.

    Returns:
        Quantity: Density [kg/m**3].

    """
    if R is None:
        R = constants.gas_constant_dry_air

    # SI conversion
    p = p.to('pascal')
    T = T.to('kelvin')
    R = R.to('joule / kelvin / kilogram')

    ret = physics.thermodynamics.density(
              p.magnitude, T.magnitude, R.magnitude)

    return ret * ureg('kilogram / meter**3')



"""Special test case which tries to load all files in ARTS XML data."""
import os

import pytest

from typhon.arts import xml
from typhon import environment


def collect_xml_files():
    """Collect all XML files in the ARTS XML data tree."""
    for d in environment.ARTS_DATA_PATH.split(os.path.pathsep):
        for root, _, filenames in os.walk(d):
            for filename in filenames:
                if filename.endswith(('.xml', '.xml.gz')):
                    yield os.path.join(root, filename)


@pytest.mark.slow
@pytest.mark.skipif(environment.ARTS_DATA_PATH is None,
                    reason='ARTS_DATA_PATH not set.')
@pytest.mark.parametrize('xmlfile', collect_xml_files())
def test_load_arts_xml_data(xmlfile):
    """Try to load all XML files in ARTS_DATA_PATH.

    Search for XML files in ARTS_DATA_PATH. If files are found, try to load
    them. It is just checked, if xml.load runs without exception.
    """
    xml.load(xmlfile)
    pass



# -*- coding: utf-8 -*-

"""This module contains functions for python's datetime/timedelta objects
"""
import functools
import time
from datetime import datetime, timedelta
from numbers import Number

import netCDF4
import numpy as np
import pandas as pd

__all__ = [
    "date2num",
    "num2date",
    "set_time_resolution",
    "to_datetime",
    "to_timedelta",
    "Timer",
]


def set_time_resolution(datetime_obj, resolution):
    """Set the resolution of a python datetime object.

    Args:
        datetime_obj: A python datetime object.
        resolution: A string indicating the required resolution.

    Returns:
        A datetime object truncated to *resolution*.

    Examples:

    .. code-block:: python

        from typhon.utils.time import set_time_resolution, to_datetime

        dt = to_datetime("2017-12-04 12:00:00")
        # datetime.datetime(2017, 12, 4, 12, 0)

        new_dt = set_time_resolution(dt, "day")
        # datetime.datetime(2017, 12, 4, 0, 0)

        new_dt = set_time_resolution(dt, "month")
        # datetime.datetime(2017, 12, 1, 0, 0)
    """
    if resolution == "year":
        return set_time_resolution(datetime_obj, "day").replace(month=1, day=1)
    elif resolution == "month":
        return set_time_resolution(datetime_obj, "day").replace(day=1)
    elif resolution == "day":
        return datetime_obj.replace(hour=0, minute=0, second=0, microsecond=0)
    elif resolution == "hour":
        return datetime_obj.replace(minute=0, second=0, microsecond=0)
    elif resolution == "minute":
        return datetime_obj.replace(second=0, microsecond=0)
    elif resolution == "second":
        return datetime_obj.replace(microsecond=0)
    elif resolution == "millisecond":
        return datetime_obj.replace(
            microsecond=int(datetime_obj.microsecond / 1000) * 1000
        )
    else:
        raise ValueError("Cannot set resolution to '%s'!" % resolution)


def to_datetime(obj):
    """Convert an object to a python datetime object.

    Args:
        obj: Can be a string with time information, a numpy.datetime64 or a
            pandas.Timestamp object.

    Returns:
        A python datetime object.

    Examples:

    .. code-block:: python

        dt = to_datetime("2017-12-04 12:00:00")
        # dt is datetime.datetime(2017, 12, 4, 12, 0)
    """

    if isinstance(obj, datetime):
        return obj
    else:
        return pd.to_datetime(obj).to_pydatetime()


def to_timedelta(obj, numbers_as=None):
    """Convert an object to a python datetime object.

    Args:
        obj: Can be a string with time information, a number, a
            numpy.timedelta64 or a pandas.Timedelta object.
        numbers_as: A string that indicates how numbers should be
            interpreted. Allowed values are *weeks*, *days*, *hours*,
            *minutes*, *seconds*, *milliseconds* and *microseconds.

    Returns:
        A python datetime object.

    Examples:

    .. code-block:: python

        # A timedelta object with 200 seconds
        t = to_timedelta("200 seconds")

        # A timedelta object with 24 days
        t = to_timedelta(24, numbers_as="days")
    """

    if numbers_as is None:
        numbers_as = "seconds"

    if isinstance(obj, timedelta):
        return obj
    elif isinstance(obj, Number):
        return timedelta(**{numbers_as: int(obj)})
    else:
        return pd.to_timedelta(obj).to_pytimedelta()


unit_mapper = {
    "nanoseconds": "ns",
    "microseconds": "us",
    "milliseconds": "ms",
    "seconds": "s",
    "hours": "h",
    "minutes": "m",
    "days": "D",
}


class InvalidUnitString(Exception):
    def __init__(self, *args, **kwargs):
        super(InvalidUnitString, self).__init__(*args, **kwargs)


def date2num(dates, units, calendar=None):
    """Convert an array of integer into datetime objects.

    This function optimizes the date2num function of python-netCDF4 if the
    standard calendar is used.

    Args:
        dates: Either an array of numpy.datetime64 objects (if standard
            gregorian calendar is used), otherwise an array of python
            datetime objects.
        units: A string with the format "{unit} since {epoch}",
            e.g. "seconds since 1970-01-01T00:00:00".
        calendar: (optional) Standard is gregorian. If others are used,
            netCDF4.num2date will be called.

    Returns:
        An array of integers.
    """
    if calendar is None:
        calendar = "gregorian"
    else:
        calendar = calendar.lower()

    if calendar != "gregorian":
        return netCDF4.date2num(dates, units, calendar)

    try:
        unit, epoch = units.split(" since ")
    except ValueError:
        raise InvalidUnitString("Could not convert to numeric values!")

    converted_data = \
        dates.astype("M8[%s]" % unit_mapper[unit]).astype("int")

    # numpy.datetime64 cannot read certain time formats while pandas can.
    epoch = pd.Timestamp(epoch).to_datetime64()

    if epoch != np.datetime64("1970-01-01"):
        converted_data -= np.datetime64("1970-01-01") - epoch
    return converted_data


def num2date(times, units, calendar=None):
    """Convert an array of integers into datetime objects.

    This function optimizes the num2date function of python-netCDF4 if the
    standard calendar is used.

    Args:
        times: An array of integers representing timestamps.
        units: A string with the format "{unit} since {epoch}",
            e.g. "seconds since 1970-01-01T00:00:00".
        calendar: (optional) Standard is gregorian. If others are used,
            netCDF4.num2date will be called.

    Returns:
        Either an array of numpy.datetime64 objects (if standard gregorian
        calendar is used), otherwise an array of python datetime objects.
    """
    try:
        unit, epoch = units.split(" since ")
    except ValueError:
        raise InvalidUnitString("Could not convert to datetimes!")

    if calendar is None:
        calendar = "gregorian"
    else:
        calendar = calendar.lower()

    if calendar != "gregorian":
        return netCDF4.num2date(times, units, calendar).astype(
            "M8[%s]" % unit_mapper[unit])

    # Numpy uses the epoch 1970-01-01 natively.
    converted_data = times.astype("M8[%s]" % unit_mapper[unit])

    # numpy.datetime64 cannot read certain time formats while pandas can.
    epoch = pd.Timestamp(epoch).to_datetime64()

    # Maybe there is another epoch used?
    if epoch != np.datetime64("1970-01-01"):
        converted_data -= np.datetime64("1970-01-01") - epoch
    return converted_data


class Timer:
    """Provide a simple time profiling utility

    Parameters:
        verbose (bool): Print measured duration after stopping the timer.
        info (str): Allows to add additional information to output.
            The given string is printed before the measured time.
            If `None`, default information is added depending on the use case.

    Returns:
        datetime.timedelta: The duration between start and end time.

    Examples:
        Timer in with statement:

        >>> import time
        >>> with Timer():
        ...     time.sleep(1)
        elapsed time: 0:00:01.003186

        Timer as object (allows to store :class:`datetime.timedelta`):

        >>> import time
        >>> t = Timer().start()
        >>> time.sleep(1)
        >>> dt = t.stop()
        elapsed time: 0:00:01.004756

        As function decorator:

        >>> @Timer()
        ... def own_function(s):
        ...     import time
        ...     time.sleep(s)
        >>> own_function(1)
        own_function: 0:00:01.004667

        Use it in format strings:

        >>> from typhon.utils import Timer
        >>> timer = Timer().start()
        >>> print(f"{timer} elapsed")
        0:00:00.000111 hours elapsed
    """
    def __init__(self, info=None, verbose=True):
        """Create a timer object."""
        self.verbose = verbose
        self.info = info

        self.starttime = None
        self.endtime = None

    def __call__(self, func):
        """Allows to use a Timer object as a decorator."""
        # When no additional information is given, add the function name if
        # Timer is used as decorator.
        if self.info is None:
            self.info = func.__name__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                # Call the original function in a Timer context.
                return func(*args, **kwargs)

        return wrapper

    def __enter__(self):
        return self.start()

    def __exit__(self, *args):
        self.stop()

    def __repr__(self):
        return repr(self.elapsed)

    def __str__(self):
        return str(self.elapsed) + " hours"

    @property
    def elapsed(self):
        """Get the elapsed time as timedelta object"""
        if self.starttime is None:
            raise ValueError("Timer has not been started yet!")

        return timedelta(seconds=time.time() - self.starttime)

    def start(self):
        """Start timer."""
        self.starttime = time.time()
        return self

    def stop(self):
        """Stop timer and print info message

        The info message will be only printed if `Timer.verbose` is *true*.

        Returns:
            A timedelta object.
        """
        if self.starttime is None:
            raise ValueError("Timer has not been started yet!")

        self.endtime = time.time()

        dt = timedelta(seconds=self.endtime - self.starttime)

        # If no additional information is specified add default information
        # to make the output more readable.
        if self.info is None:
            self.info = 'elapsed time'

        if self.verbose:
            # Connect additional information and measured time for output.
            print('{info}: {time}'.format(info=self.info, time=dt))

        return dt



""" The variables submodule.

This module contains symbolic representations of all ARTS workspace variables.

The variables are loaded dynamically when the module is imported, which ensures that they
up to date with the current ARTS build.

TODO: The group names list is redudant w.rt. group_ids.keys(). Should be removed.

Attributes:
    group_names([str]): List of strings holding the groups of ARTS WSV variables.
    group_ids(dict):    Dictionary mapping group names to the group IDs which identify
                        groups in the ARTS C API.
"""

import ctypes as c
import os
import numpy as np
import re
import scipy as sp
import tempfile

from typhon.arts.workspace.api import arts_api
from typhon.arts.workspace.agendas import Agenda
from typhon.arts.xml.names import tensor_names


class WorkspaceVariable:
    """
    The WorkspaceVariable represents ARTS workspace variables in a symbolic way. This
    means that they are not associated with a single workspace and therefore do not have a
    unique value. Their value in a given workspacecan be accessed, however, using the value()
    method.

    Attributes:
    ws_id(int):          The Index variable identifying the variable in the ARTS C API.
    name(str):        The name of the workspace variable.
    group(str):       The name of the group this variable belongs to.
    description(str): The documentation of the variable as in methods.cc
    """
    def __init__(self, ws_id, name, group, description, ws = None):
        self.ws_id = ws_id
        self.name = name
        self.group = group
        self.group_id = group_ids[group]
        self.description = description
        self.ws = ws

        self.ndim = None
        if self.group == "Vector":
            self.ndim = 1
        if self.group == "Matrix":
            self.ndim = 2
        m = re.match(r"^Tensor(\d)$", self.group)
        if m:
            self.ndim = int(m.group(1))

        self.update()

    def __getstate__(self):
        return self.ws_id, self.name, self.group, \
               self.group_id, self.description, self.ndim

    def __setstate__(self, state):
        self.ws_id, self.name, self.group, self.group_id, self.description,\
            self.ndim = state

    def __repr__(self):
        s  = "ARTS Workspace Variable\n\n"
        s += "Name:  " + self.name + "\n"
        s += "Group: " + self.group + "\n\n"
        s += self.description
        return s

    def __str__(self):
        return self.__repr__()

    def __setattr__(self, name, value):

        if name == "value":
            if self.ws is None:
                raise Exception("Cannot set value of WSV without associated "
                                " workspace.")
            else:
                self.ws.__setattr__(self.name, value)
        else:
            super().__setattr__(name, value)

    def print(self):
        """ Print variable value using ARTS Print(...) WSM.

        Raises:
            Exception: If the variable has no associated workspace.
        """
        if (self.ws):
            self.ws.Print(self, 1)
        else:
            raise Exception("Can't print variable without associated ARTS workspace.")

    @staticmethod
    def get_variable_name(i):
        """
        Lookup the name of a variable given its workspace index.

        Args:
            i(int): The index of the workspace variable.

        Returns:
            str: The name of the workspace variable.
        """
        s = arts_api.get_variable(i)
        name = s.name.decode("utf8")
        return name

    @staticmethod
    def get_group_id(value):
        """ This static method is used to determine how (and if) a given python variable can
        be mapped to a ARTS workspace variable group. The returned group id is required to
        add the variable to a workspace.

        Args:
            value(any): The python variable to map to the ARTS group.

        Returns:
            int: The index of the group which can be used to represent the python variable
                 or None if the type is not supported.
        """
        if isinstance(value, WorkspaceVariable):
            return group_ids[value.group]
        elif isinstance(value, Agenda):
            return group_ids["Agenda"]
        elif isinstance(value, np.long):
            return group_ids["Index"]
        elif isinstance(value, (float, np.double)):
            return group_ids["Numeric"]
        elif isinstance(value, str):
            return group_ids["String"]
        elif isinstance(value, np.ndarray):
            if value.ndim == 1:
                return group_ids["Vector"]
            elif value.ndim == 2:
                return group_ids["Matrix"]
            elif value.ndim == 3:
                return group_ids["Tensor3"]
            elif value.ndim == 4:
                return group_ids["Tensor4"]
            elif value.ndim == 5:
                return group_ids["Tensor5"]
            elif value.ndim == 6:
                return group_ids["Tensor6"]
            elif value.ndim == 7:
                return group_ids["Tensor7"]
            else:
                raise ValueError(
                    "Numpy arrays are only supported up to 7 dimensions."
                )
        elif sp.sparse.issparse(value):
            return group_ids["Sparse"]
        elif type(value) == list:
            group_name = ""
            nested_value = value
            while type(nested_value) == list and len(nested_value) > 0:
                nested_value = nested_value[0]
                group_name += "ArrayOf"
            if type(nested_value) == list and len(nested_value) == 0:
                raise ValueError("Empty lists are currently not handled.")
            else:
                typename = type(nested_value).__name__
                if isinstance(nested_value, str):
                    group_name += "String"
                    return group_ids[group_name]
                elif isinstance(nested_value, np.long):
                    group_name += "Index"
                    return group_ids[group_name]
                elif isinstance(nested_value, (float, np.double)):
                    raise ValueError("Vectors, Matrices or Tensors should be"
                                     " passed as numpy.ndarray and not as"
                                     " lists.")
                elif hasattr(nested_value, 'write_xml') and typename in group_names:
                    return group_ids[group_name + typename]
                elif isinstance(nested_value, np.ndarray):
                    group_name += tensor_names[len(nested_value.shape) - 1]
                    return group_ids[group_name]
                else:
                    raise ValueError(
                        f"Nested array with internal type "
                        f"{type(nested_value)} not supported.")
        elif hasattr(value, 'write_xml') and type(value).__name__ in group_names:
            return group_ids[type(value).__name__]
        else:
            raise ValueError(f"Type {type(value)} currently not supported.")

    @classmethod
    def convert(cls, group, value):
        """ Tries to convert a given python object to an object of the python class
        representing the given ARTS WSV group.

        Args:
            group(string): The name of an ARTS WSV group.
            group(any):    The object to convert

        Returns:
            (any): The converted object.
        """
        if (group == "Index"):
            return int(value)
        if (group == "String"):
            return value
        if (group == "ArrayOfString"):
            return [str(i) for i in value]
        if (group == "Numeric"):
            return np.float64(value)
        if (group == "Vector"):
            return np.array(value, dtype=np.float64, order='C', ndmin=1)
        if (group == "Matrix"):
            return np.array(value, dtype=np.float64, order='C', ndmin=2)
        if (group == "Sparse"):
            return sp.sparse.coo_matrix(value)
        if (group[:6] == "Tensor"):
            dim = int(group[6])
            return np.array(value, dtype=np.float64, order='C', ndmin=dim)
        if group.startswith("ArrayOf"):
            subgroup = group[7:]
            if hasattr(value, "__iter__"):
                return [cls.convert(subgroup, v) for v in value]
            else:
                return [cls.convert(subgroup, value)]
        return None

    @staticmethod
    def iter():
        """
        Iterator returning a WorkspaceVariable object for each ARTS WSV available.
        """
        for i in range(arts_api.get_number_of_variables()):
            s = arts_api.get_variable(i)
            name        = s.name.decode("utf8")
            description = s.description.decode("utf")
            group       = group_names[s.group]
            yield WorkspaceVariable(i, name, group, description)

    @property
    def initialized(self):

        ws = self.ws
        if ws is None:
            raise ValueError("WorkspaceVariable object needs associated"
                             " Workspace to determine value.")

        v = arts_api.get_variable_value(ws.ptr, self.ws_id, self.group_id)
        return v.initialized

    @property
    def value(self):
        """ Return the value of the variable in a given workspace.

        By default this function will check the value in the workspace associated
        with the variable of in the workspace object provided as argument to the
        function call. If the variable has an associated workspace the workspace
        provided as argument will be ignored.

        Returns:
            The value of the workspace variable represented by an object of
            the corresponding python types.

        Raises:
            Exception: If the type of the workspace variable is not supported
            by the interface.

        """
        from typhon.arts.types import classes as typhon_classes


        if (self.ws):
            ws = self.ws
        if not ws:
            raise ValueError("WorkspaceVariable object need Workspace to determine value.")

        v = arts_api.get_variable_value(ws.ptr, self.ws_id, self.group_id)
        if not v.initialized:
            raise Exception("WorkspaceVariable " + self.name + " is uninitialized.")

        if self.group in typhon_classes:
            cls = typhon_classes[self.group]
            if hasattr(cls, "__from_variable_value_struct__"):
                return cls.__from_variable_value_struct__(v)
        if self.group == "Index":
            return c.cast(v.ptr, c.POINTER(c.c_long))[0]
        elif self.group == "Numeric":
            return c.cast(v.ptr, c.POINTER(c.c_double))[0]
        elif self.group == "String":
            return (c.cast(v.ptr, c.c_char_p)).value.decode("utf8")
        elif self.group == "ArrayOfIndex":
            return [c.cast(v.ptr, c.POINTER(c.c_long))[i]
                    for i in range(v.dimensions[0])]
        elif self.group == "Sparse":
            m    = v.dimensions[0]
            n    = v.dimensions[1]
            nnz  = v.dimensions[2]
            if nnz == 0:
                return sp.sparse.csr_matrix(0)
            else:
                print(m, n, nnz)
                data = np.ctypeslib.as_array(c.cast(v.ptr,
                                                    c.POINTER(c.c_double)),
                                             (nnz,))
                row_indices = np.ctypeslib.as_array(v.inner_ptr, (nnz,))
                col_starts  = np.ctypeslib.as_array(v.outer_ptr, (m + 1,))
                return sp.sparse.csr_matrix((data, row_indices, col_starts),
                                            shape=(m,n))
        elif self.group == "Agenda":
            return Agenda(v.ptr)
        elif self.ndim:
            shape = []
            size  = 1
            for i in range(self.ndim):
                shape.append(v.dimensions[i])
                size *= v.dimensions[i]
            if size > 0:
                self.__array_interface__ = {"shape"  : tuple(shape),
                                            "typestr" : "|f8",
                                            "data" : (v.ptr, False),
                                            "version" : 3}
                return np.asarray(self)
            else:
                return np.zeros(shape)
        else:
            try:
                return self.to_typhon()
            except:
                raise Exception("Type of workspace variable is not supported "
                                + " by the interface.")

    def update(self):
        """ Update data references of the object.

        References to vector, matrices and tensors may change and must therefore
        be updated dynamically to ensure they are consistent with the state of
        the associated workspace. This method takes care of that.

        """
        if not self.ws==None and self.ndim:
            v = arts_api.get_variable_value(self.ws.ptr, self.ws_id, self.group_id)
            shape = []
            for i in range(self.ndim):
                shape.append(v.dimensions[i])
            self.__array_interface__ = {"shape"  : tuple(shape),
                                        "typestr" : "|f8",
                                        "data" : (v.ptr, False),
                                        "version" : 3}

    def erase(self):
        """
        Erase workspace variable from its associated workspace.
        """
        if self.ws:
            arts_api.erase_variable(self.ws.ptr, self.ws_id, self.group_id)
            self.ws = None

    def describe(self):
        """
        Print the description of the variable as given in ARTS methods.cc
        """
        print(self.description.format())

    def to_typhon(self):
        """
        Return the value of this variable as a typhon type. This function
        writes the value of the variable to a temporary file and reads it
        into Python using typhon load function. The purpose of this function
        is to access WSV whose groups are not natively supported by the
        C API.

        Returns:
            A typhon object with the same value as the WSV in the associated
            workspace.
        """
        from typhon.arts.xml import load

        if not self.ws:
            raise Exception("Cannot retrieve the value of a variable without "
                            + " associated Workspace.")
        with tempfile.TemporaryDirectory() as tmpdir:
            tfile = os.path.join(tmpdir, 'wsv.xml')
            self.ws.WriteXML("binary", self, tfile)
            v = load(tfile)

        return v

    def from_typhon(self, var):
        """
        Set the value of this WSV in the associated workspace to the given
        typhon type. This function writes the value in ASCII format to a
        temporary file and reads it into the workspace

        Args:
            var: The value to which this WSV should be set in the associated
                 workspace.

        """
        from typhon.arts.xml import save

        if not self.ws:
            raise Exception("Cannot set the value of a variable without "
                            + " associated Workspace.")
        with tempfile.TemporaryDirectory() as tmpdir:
            tfile = os.path.join(tmpdir, 'wsv.xml')
            save(var, tfile, format='binary')
            self.ws.ReadXML(self, tfile)


# Get ARTS WSV groups
group_names = [arts_api.get_group_name(i).decode("utf8")
               for i in range(arts_api.get_number_of_groups())]
group_ids = dict([(id, name) for (name,id) in enumerate(group_names)])


workspace_variables = dict()
for v in WorkspaceVariable.iter():
    globals()[v.name] = v
    workspace_variables[v.name] = v




# -*- coding: utf-8 -*-
"""
Created on Thu Jul  6 17:25:02 2017

@author: yamada
"""
import numpy as np
from scipy.special import wofz
from scipy.constants import c, k, R
# =============================================================================
# h =6.62607004*10**-34  # Js or mmkg/s
# k = 1.38064852*10**-23  # mmkg/ssK  = J/K
# c = 299792458.  # m/s
#ac = 6.02214086e23  # Avogadro constant/M
from scipy.constants import Avogadro as ac
#6.022140857e+23
ac = 6.0220450e26  # Avogadro constant/M
# debye = 3.33564e-30  # 1D = 3.34e-30 [Cm]
# R = 8.3144598  # J/K/mol universal gas constant
# =============================================================================


def DLV(Type, gct, *,Freq=0, gcp=1, gcv=1, Para=1, HWHM=False):
    if Type == 'D':  # dopplar
        u"""#doppler width
        #Para[transient Freq[Hz], relative molecular mass[g/mol]]"""
        # step1 = Para[0]/c*(2.*R*gct/(Para[1]*1.e-3))**0.5
        # outy = np.exp(-(Freq-Para[0])**2/step1**2) / (step1*(np.pi**0.5))
        step1 = Para[0]/c*(2.*R*gct*np.log(2.)/(Para[1]*1.e-3))**0.5  # HWHM
        outy = np.exp(-np.log(2.)*(Freq-Para[0])**2/step1**2) *\
                     (np.log(2.)/np.pi)**0.5/step1
#        GD = np.sqrt(2*k*ac/18.0153*gct)/c*Para[0]
#        step1 = GD
#        outy = wofz((Freq-Para[0])/GD).real / np.sqrt(np.pi) / GD
#        print(outy.shape)
    elif Type == 'L':  # Lorenz
        """#Collisional width
        Temperature [K]
        pressure [atm]: 1atm = 101325Pa
        Number density of a specie [m-3]
        Para[transient Freq,gamm air, gamma self ,air n]
        gamma [cm-1 atm-1]"""
        vmrs = gcv/(gcp/k/gct)  # gamma
        step1 = (Para[1]*gcp/101325.*(1.-vmrs)+Para[2]*gcp/101325.*(vmrs)) *\
                (296./gct)**Para[3]*100*c  # cm-1*100*c ->s-1
        # step1 = (Para[1]*gcp/101325.*(1.-vmrs))*(296./gct)**Para[3]*100*c
        outy = step1/(np.pi*((Freq-Para[0])**2 + step1**2))
    elif Type == 'V':  # Voigt
        """#Para[frequency[Jup]*1.e9,
        gamma_air[Hn],gamma_self[Hn],n_air[Hn],M]#"""
        ad = Para[0]/c*(2.*R*gct*np.log(2.)/(Para[4]*1.e-3))**0.5  # vd
        # fd = np.exp(-np.log(2.)*(Freq-Para[0])**2/ad**2) *\
        #            (np.log(2.)/np.pi)**0.5/ad  # #fd
        vmrs = gcv/(gcp/k/gct)  # gamma
        al = (Para[1]*gcp/101325.*(1.-vmrs)) *\
             (296./gct)**Para[3]*100*c  # alphaL
        # fl = ad/(np.pi*((Freq-Para[0])**2 + ad**2))  #
        # outy = np.cumsum(fv1,axis=1)
        # wofz_y = al/ad  # see (2.46)
        # wofz_x = (Freq-Para[0])/ad  # see (2.46)
        sigma = ad/(2.*np.log(2))**0.5
        gamma = al
        for iii in range(al.size):
            # z = wofz_y[iii] - ( wofz_x[:,iii]*1j )
            # Z = np.array(wofz(z),dtype = complex)/ad[iii]/np.pi**0.5
            # wofz:exp(z**2)*ergc(z)#ergc:error_function
            if iii == 0:
                outy = np.real(wofz((Freq.reshape(Freq.size)-Para[0] +
                                     1j*gamma[iii])/sigma[iii]/2**0.5)) /\
                                     sigma[iii]/(2*np.pi)**0.5
            else:
                outy = np.vstack((outy,
                                  np.real(wofz((Freq.reshape(Freq.size) -
                                                Para[0] + 1j*gamma[iii]) /
                                               sigma[iii]/2**0.5)) /
                                 sigma[iii]/(2*np.pi)**0.5))
        outy = outy.T
        # outy = np.real(Z)
    else:
        raise ValueError('Do you wanna calculate other shape function?')
    if HWHM is True:
        return step1, outy
    elif HWHM is False:
        return outy


def Linewidth(Type, gct, Para):
    """#doppler width
    #Para[transient Freq[Hz], relative molecular mass[g/mol]]"""
    step1 = Para[0]/c*(2.*R*gct/(Para[1]*1.e-3))**0.5
    step1 = Para[0]/c*(2.*R*gct*np.log(2.)/(Para[1]*1.e-3))**0.5  # HWHM
    return int(step1.max())



def DopplerWind(Temp, FreqGrid, Para, wind_v, shift_direction='red'):
    u"""#doppler width
    #Para[transient Freq[Hz], relative molecular mass[g/mol]]"""
    # step1 = Para[0]/c*(2.*R*gct/(Para[1]*1.e-3))**0.5
    # outy = np.exp(-(Freq-Para[0])**2/step1**2) / (step1*(np.pi**0.5))
    #wind_v = speed[:,10] 
    #Temp=temp[10]
    #FreqGrid = Fre_range_i[0]
    wind = wind_v.reshape(wind_v.size, 1)
    FreqGrid = FreqGrid.reshape(1, FreqGrid.size)
    deltav = Para[0]*wind/c
    if shift_direction.lower() == 'red':
        D_effect = (deltav)
    elif shift_direction.lower() == 'blue':
        D_effect = (-deltav)
    else:
        raise ValueError('Set shift direction to "red" or "blue".')

#    step1 = Para[0]/c*(2.*R*Temp*np.log(2.)/(Para[1]*1.e-3))**0.5  # HWHM
#    outy = np.exp(-np.log(2.)*(FreqGrid-Para[0])**2/step1**2) *\
#                 (np.log(2.)/np.pi)**0.5/step1
#    outy_d = np.exp(-np.log(2.)*(FreqGrid+D_effect-Para[0])**2/step1**2) *\
#                   (np.log(2.)/np.pi)**0.5/step1
    GD = np.sqrt(2*k*ac/Para[1]*Temp)/c*Para[0]
    step1 = GD
    outy_d = wofz((FreqGrid+D_effect-Para[0])/GD).real / np.sqrt(np.pi) / GD
    #plot(FreqGrid, outy)
    #plot(FreqGrid, outy_d[:,0])
    return outy_d








# -*- coding: utf-8 -*-
"""
This module provides colormaps to use for the visualisation of meteorological
data.  It heavily bases on the cmocean_ package developed by Kristen Thyng.
Most colormaps are directly inherited and renamed for meteorological
applications.

The colormaps are registered in matplotlib after importing typhon:

    >>> import typhon
    >>> plt.get_cmap('difference')

.. _cmocean: http://matplotlib.org/cmocean/
"""
from matplotlib.colors import (LinearSegmentedColormap, ListedColormap)
from matplotlib.cm import register_cmap

from ._cmocean import datad as _cmocean_datad
from ._cm import datad as _cm_datad


datad = _cmocean_datad
datad.update(_cm_datad)


def _rev_cdict(cdict):
    """Revert a dictionary containing specs for a LinearSegmentedColormap."""
    rev_cdict = {}
    for k, v in cdict.items():
        rev_cdict[k] = [(1.0 - x, y1, y0) for x, y0, y1 in reversed(v)]
    return rev_cdict

cmaps = {}
for (name, data) in datad.items():
    if isinstance(data, dict):
        cmaps[name] = LinearSegmentedColormap(name, data)
        cmaps[name + '_r'] = LinearSegmentedColormap(
            name + '_r', _rev_cdict(data))
    elif isinstance(data, list):
        cmaps[name] = ListedColormap(data, name)
        cmaps[name + '_r'] = ListedColormap(data[::-1], name + '_r')

locals().update(cmaps)

for name, cm in cmaps.items():
    register_cmap(name, cm)

__all__ = list(cmaps.keys())



"""Utilities related to caching and memoisation
"""

# Any commits made to this module between 2015-05-01 and 2017-03-01
# by Gerrit Holl are developed for the EC project “Fidelity and
# Uncertainty in Climate Data Records from Earth Observations (FIDUCEO)”.
# Grant agreement: 638822
# 
# All those contributions are dual-licensed under the MIT license for use
# in typhon, and the GNU General Public License version 3.

import functools
import logging
import copy


__all__ = [
    'mutable_cache',
]


logger = logging.getLogger(__name__)
def mutable_cache(maxsize=10):
    """In-memory cache like functools.lru_cache but for any object

    This is a re-implementation of functools.lru_cache.  Unlike
    functools.lru_cache, it works for any objects, mutable or not.
    Therefore, it returns returns a copy and it is wrong if the mutable
    object has changed!  Use with caution!

    If you call the *resulting* function with a keyword argument
    'CLEAR_CACHE', the cache will be cleared.  Otherwise, cache is rotated
    when more than `maxsize` elements exist in the cache.  Additionally,
    if you call the resulting function with NO_CACHE=True, it doesn't
    cache at all.  Be careful with functions returning large objects, such
    as reading routines for datasets like IASI.  Everything is kept in RAM!

    Args:
        maxsize (int): Maximum number of return values to be remembered.

    Returns:
        New function that has caching implemented.
    """

    sentinel = object()
    make_key = functools._make_key

    def decorating_function(user_function):
        cache = {}
        cache_get = cache.get
        keylist = []  # don't make it too long

        def wrapper(*args, **kwds):
            if kwds.get("CLEAR_CACHE"):
                del kwds["CLEAR_CACHE"]
                cache.clear()
                keylist.clear()
            if kwds.get("NO_CACHE"):
                del kwds["NO_CACHE"]
                return user_function(*args, **kwds)
            elif "NO_CACHE" in kwds:
                del kwds["NO_CACHE"]
            # Problem with pickle: dataset objects (commonly passed as
            # 'self') contain a cache which is a shelve object which
            # cannot be pickled.  Would need to create a proper pickle
            # protocol for dataset objects... maybe later
            # key = pickle.dumps(args, 1) + pickle.dumps(kwds, 1)
            key = str(args) + str(kwds)
            result = cache_get(key, sentinel)
            if result is not sentinel:
                logger.debug(("Getting result for {!s} from cache "
                               " (key {!s}").format(user_function, key))
                # make sure we return a copy of the result; when a = f();
                # b = f(), users should reasonably expect that a is not b.
                return copy.copy(result)
#            logger.debug("No result in cache")
            result = user_function(*args, **kwds)
#            logger.debug("Storing result in cache")
            cache[key] = result
            keylist.append(key)
            if len(keylist) > maxsize:
                try:
                    del cache[keylist[0]]
                    del keylist[0]
                except KeyError:
                    pass
            return result

        return functools.update_wrapper(wrapper, user_function)

    return decorating_function
